import re
import time
from lxml import etree
import select
import socket
import logging
import sys

from lxml.builder import E
from lxml.etree import XMLSyntaxError
from datetime import datetime, timedelta
from ncclient.operations.rpc import RPCReply, RPCError
from ncclient.xml_ import to_ele
import six
from ncclient.transport.session import HelloHandler


class PY6:
    NEW_LINE = six.b("\n")
    EMPTY_STR = six.b("")
    NETCONF_EOM = six.b("]]>]]>")
    STARTS_WITH = six.b("<!--")


__all__ = ["xmlmode_netconf"]

_NETCONF_EOM = six.b("]]>]]>")
_xmlns = re.compile(six.b("xmlns=[^>]+"))
_xmlns_strip = lambda text: _xmlns.sub(PY6.EMPTY_STR, text)
_junosns = re.compile(six.b("junos:"))
_junosns_strip = lambda text: _junosns.sub(PY6.EMPTY_STR, text)

logger = logging.getLogger("jnpr.junos.tty_netconf")

# =========================================================================
# xmlmode_netconf
# =========================================================================


class tty_netconf(object):

    """
    Basic Junos XML API for bootstraping through the TTY
    """

    def __init__(self, tty):
        self._tty = tty
        self.hello = None
        self._session_id = -1

    # -------------------------------------------------------------------------
    # NETCONF session open and close
    # -------------------------------------------------------------------------

    def open(self, at_shell):
        """start the XML API process and receive the 'hello' message"""
        nc_cmd = ("junoscript", "xml-mode")[at_shell]
        self._tty.write(nc_cmd + " netconf need-trailer")
        mark_start = datetime.now()
        mark_end = mark_start + timedelta(seconds=15)

        while datetime.now() < mark_end:
            time.sleep(0.1)
            line = self._tty.read()
            if line.startswith(PY6.STARTS_WITH):
                break
        else:
            # exceeded the while loop timeout
            raise RuntimeError("Error: netconf not responding")

        self.hello = self._receive()
        self._session_id, _ = HelloHandler.parse(self.hello.decode("utf-8"))

    def close(self, force=False):
        """issue the XML API to close the session"""

        # if we do not have an open connection, then return now.
        if force is False:
            if self.hello is None:
                return

        self.rpc("close-session")
        # removed flush

    # -------------------------------------------------------------------------
    # MISC device commands
    # -------------------------------------------------------------------------

    def zeroize(self):
        """issue a reboot to the device"""
        cmd = E.command("request system zeroize")
        try:
            encode = None if sys.version < "3" else "unicode"
            self.rpc(etree.tostring(cmd, encoding=encode))
        except:
            return False
        return True

    # -------------------------------------------------------------------------
    # XML RPC command execution
    # -------------------------------------------------------------------------

    def rpc(self, cmd):
        """
        Write the XML cmd and return the response as XML object.

        :cmd:
          <str> of the XML command.  if the :cmd: is not XML, then
          this routine will perform the brackets; i.e. if given
          'get-software-information', this routine will turn
          it into '<get-software-information/>'

        NOTES:
          The return XML object is the first child element after
          the <rpc-reply>.  There is also no error-checking
          performing by this routine.
        """
        if not cmd.startswith("<"):
            cmd = "<{}/>".format(cmd)
        rpc = six.b("<rpc>{}</rpc>".format(cmd))
        logger.info("Calling rpc: %s" % rpc)
        self._tty.rawwrite(rpc)

        rsp = self._receive()
        rsp = rsp.decode("utf-8") if isinstance(rsp, bytes) else rsp
        reply = RPCReply(rsp, huge_tree=self._tty._huge_tree)
        errors = reply.errors
        if len(errors) > 1:
            raise RPCError(to_ele(reply._raw), errs=errors)
        elif len(errors) == 1:
            raise reply.error
        return rsp

    # -------------------------------------------------------------------------
    # LOW-LEVEL I/O for reading back XML response
    # -------------------------------------------------------------------------

    def _receive(self):
        # On windows select.select throws io.UnsupportedOperation: fileno
        # so use read function for windows serial COM ports
        if hasattr(self._tty, "port") and str(self._tty.port).startswith("COM"):
            return self._receive_serial_win()
        else:
            return self._receive_serial()

    def _receive_serial(self):
        """process the XML response into an XML object"""
        rxbuf = PY6.EMPTY_STR
        line = PY6.EMPTY_STR
        while True:
            try:
                rd, wt, err = select.select([self._tty._rx], [], [], 0.1)
            except select.error as err:
                raise err
            except socket.error as err:
                raise err
            if rd:
                line, lastline = rd[0].read_until(PY6.NETCONF_EOM, 0.1), line
                if not line:
                    continue
                if _NETCONF_EOM in line or _NETCONF_EOM in lastline + line:
                    rxbuf = rxbuf + line
                    break
                else:
                    rxbuf = rxbuf + line
                    if _NETCONF_EOM in rxbuf:
                        break
        return self._parse_buffer(rxbuf)

    # -------------------------------------------------------------------------
    # Read message from windows COM ports
    # -------------------------------------------------------------------------

    def _receive_serial_win(self):
        """process incoming data from windows port"""
        rxbuf = PY6.EMPTY_STR
        line = PY6.EMPTY_STR
        while True:
            line, lastline = self._tty.read().strip(), line
            if not line:
                continue
            if _NETCONF_EOM in line or _NETCONF_EOM in lastline + line:
                rxbuf = rxbuf + line
                break
            else:
                rxbuf = rxbuf + line
                if _NETCONF_EOM in rxbuf:
                    break
        return self._parse_buffer(rxbuf)

    def _parse_buffer(self, rxbuf):
        rxbuf = rxbuf.splitlines()
        if _NETCONF_EOM in rxbuf[-1]:
            if rxbuf[-1] == _NETCONF_EOM:
                rxbuf.pop()
            else:
                rxbuf[-1] = rxbuf[-1].split(_NETCONF_EOM)[0]
        try:
            rxbuf = [i.strip() for i in rxbuf if i.strip() != PY6.EMPTY_STR]
            rcvd_data = PY6.NEW_LINE.join(rxbuf)
            logger.debug("Received: \n%s" % rcvd_data)
            parser = etree.XMLParser(
                remove_blank_text=True, huge_tree=self._tty._huge_tree
            )
            try:
                etree.XML(rcvd_data, parser)
            except XMLSyntaxError:
                if _NETCONF_EOM in rcvd_data:
                    rcvd_data = rcvd_data[: rcvd_data.index(_NETCONF_EOM)]
                    etree.XML(rcvd_data)  # just to recheck
                else:
                    parser = etree.XMLParser(recover=True)
                    rcvd_data = etree.tostring(etree.XML(rcvd_data, parser=parser))
            return rcvd_data
        except:
            if "</xnm:error>" in rxbuf:
                for x in rxbuf:
                    if "<message>" in x:
                        return etree.XML(
                            "<error-in-receive>" + x + "</error-in-receive>"
                        )
            else:
                return etree.XML("<error-in-receive/>")



"""
This file contains the FactoryLoader class that is used to dynamically
create Runstat Table and View objects from a <dict> of data.  The <dict> can
originate from any kind of source: YAML, JSON, program.  For examples of YAML
refer to the .yml files in this jnpr.junos.op directory.
"""
# stdlib
from copy import deepcopy
import re

from jinja2 import Environment

# locally
from jnpr.junos.factory.factory_cls import *
from jnpr.junos.factory.viewfields import *

__all__ = ["FactoryLoader"]

# internally used shortcuts

_VIEW = FactoryView
_CMDVIEW = FactoryCMDView
_FIELDS = ViewFields
_GET = FactoryOpTable
_TABLE = FactoryTable
_CFGTBL = FactoryCfgTable
_CMDTBL = FactoryCMDTable
_CMDCHILDTBL = FactoryCMDChildTable


class FactoryLoader(object):

    """
    Used to load a <dict> of data that contains Table and View definitions.

    The primary method is :load(): which will return a <dict> of item-name and
    item-class definitions.

    If you want to import these definitions directly into your namespace,
    (like a module) you would do the following:

      loader = FactoryLoader()
      catalog = loader.load( <catalog_dict> )
      globals().update( catalog )

    If you did not want to do this, you can access the items as the catalog.
    For example, if your <catalog_dict> contained a Table called MyTable, then
    you could do something like:

      MyTable = catalog['MyTable']
      table = MyTable(dev)
      table.get()
      ...
    """

    def __init__(self):
        self._catalog_dict = None  # YAML data

        self._item_optables = []  # list of the get/op-tables
        self._item_cfgtables = []  # list of get/cfg-tables
        self._item_cmdtables = []  # list of commands with unstructured data o/p
        self._item_views = []  # list of views to build
        self._item_tables = []  # list of tables to build

        self.catalog = {}  # catalog of built classes

    # -----------------------------------------------------------------------
    # Create a View class from YAML definition
    # -----------------------------------------------------------------------

    def _fieldfunc_True(self, value_rhs):
        def true_test(x):
            if value_rhs.startswith("regex("):
                return True if bool(re.search(value_rhs.strip("regex()"), x)) else False
            return x == value_rhs

        return true_test

    def _fieldfunc_False(self, value_rhs):
        def false_test(x):
            if value_rhs.startswith("regex("):
                return False if bool(re.search(value_rhs.strip("regex()"), x)) else True
            return x != value_rhs

        return false_test

    def _fieldfunc_Search(self, regex_pattern):
        def search_field(field_text):
            """Returns the first occurrence of regex_pattern within given field_text."""
            match = re.search(regex_pattern, field_text)
            if match:
                return match.groups()[0]
            else:
                return None

        return search_field

    def _add_dictfield(self, fields, f_name, f_dict, kvargs):
        """add a field based on its associated dictionary"""
        # at present if a field is a <dict> then there is **one
        # item** - { the xpath value : the option control }.  typically
        # the option would be a bultin class type like 'int'
        # however, as this framework expands in capability, this
        # will be enhaced, yo!

        xpath, opt = list(f_dict.items())[0]  # get first/only key,value

        if opt == "group":
            fields.group(f_name, xpath)
            return

        if "flag" == opt:
            opt = "bool"  # flag is alias for bool

        # first check to see if the option is a built-in Python
        # type, most commonly would be 'int' for numbers, like counters
        if isinstance(opt, dict):
            kvargs.update(opt)
            fields.str(f_name, xpath, **kvargs)
            return

        astype = __builtins__.get(opt) or globals().get(opt)
        if astype is not None:
            kvargs["astype"] = astype
            fields.astype(f_name, xpath, **kvargs)
            return

        # next check to see if this is a "field-function"
        # operator in the form "func=value", like "True=enabled"

        if isinstance(opt, str) and opt.find("=") > 0:
            field_cmd, value_rhs = opt.split("=")
            fn_field = "_fieldfunc_" + field_cmd
            if not hasattr(self, fn_field):
                raise ValueError("Unknown field-func: '%'" % field_cmd)
            kvargs["astype"] = getattr(self, fn_field)(value_rhs)
            fields.astype(f_name, xpath, **kvargs)
            return

        raise RuntimeError("Dont know what to do with field: '%s'" % f_name)

    # ---[ END: _add_dictfield ] ---------------------------------------------

    def _add_view_fields(self, view_dict, fields_name, fields):
        """add a group of fields to the view"""
        fields_dict = view_dict[fields_name]
        try:
            # see if this is a 'fields_<group>' collection, and if so
            # then we automatically setup using the group mechanism
            mark = fields_name.index("_")
            group = {"group": fields_name[mark + 1 :]}
        except:
            # otherwise, no group, just standard 'fields'
            group = {}

        for f_name, f_data in fields_dict.items():
            # each field could have its own unique set of properties
            # so create a kvargs <dict> each time.  but copy in the
            # groups <dict> (single item) generically.
            kvargs = {}
            kvargs.update(group)

            if isinstance(f_data, dict):
                self._add_dictfield(fields, f_name, f_data, kvargs)
                continue

            if f_data in self._catalog_dict:
                # f_data is the table name
                cls_tbl = self.catalog.get(f_data, self._build_table(f_data))
                fields.table(f_name, cls_tbl)
                continue

            # if we are here, then it means that the field is a string value
            xpath = f_name if f_data is True else f_data
            fields.str(f_name, xpath, **kvargs)

    def _add_cmd_view_fields(self, view_dict, fields_name, fields):
        """add a group of fields to the view"""
        fields_dict = view_dict[fields_name]
        for f_name, f_data in fields_dict.items():
            if f_data in self._catalog_dict:
                cls_tbl = self.catalog.get(f_data, self._build_cmdtable(f_data))
                fields.table(f_name, cls_tbl)
                continue

            # if we are here, it means we need to filter fields from textfsm
            fields._fields.update({f_name: f_data})

    # -------------------------------------------------------------------------

    def _build_view(self, view_name):
        """build a new View definition"""
        if view_name in self.catalog:
            return self.catalog[view_name]

        view_dict = self._catalog_dict[view_name]
        kvargs = {"view_name": view_name}

        # if there are field groups, then get that now.
        if "groups" in view_dict:
            kvargs["groups"] = view_dict["groups"]

        # if there are eval, then get that now.
        if "eval" in view_dict:
            kvargs["eval"] = {}
            for key, exp in view_dict["eval"].items():
                env = Environment()
                kvargs["eval"][key] = env.parse(exp)

        # if this view extends another ...
        if "extends" in view_dict:
            base_cls = self.catalog.get(view_dict["extends"])
            # @@@ should check for base_cls is None!
            kvargs["extends"] = base_cls

        fields = _FIELDS()
        fg_list = [name for name in view_dict if name.startswith("fields")]
        for fg_name in fg_list:
            self._add_view_fields(view_dict, fg_name, fields)

        cls = _VIEW(fields.end, **kvargs)
        self.catalog[view_name] = cls
        return cls

    # -------------------------------------------------------------------------

    def _build_cmdview(self, view_name):
        """build a new View definition"""
        if view_name in self.catalog:
            return self.catalog[view_name]

        view_dict = self._catalog_dict[view_name]
        kvargs = {"view_name": view_name}

        if "columns" in view_dict:
            kvargs["columns"] = view_dict["columns"]
        elif "title" in view_dict:
            kvargs["title"] = view_dict["title"]
        if "regex" in view_dict:
            kvargs["regex"] = view_dict["regex"]
        if "exists" in view_dict:
            kvargs["exists"] = view_dict["exists"]
        if "filters" in view_dict:
            kvargs["filters"] = view_dict["filters"]
        if "eval" in view_dict:
            kvargs["eval"] = {}
            for key, exp in view_dict["eval"].items():
                env = Environment()
                kvargs["eval"][key] = env.parse(exp)
        fields = _FIELDS()
        fg_list = [name for name in view_dict if name.startswith("fields")]
        for fg_name in fg_list:
            self._add_cmd_view_fields(view_dict, fg_name, fields)

        cls = _CMDVIEW(fields.end, **kvargs)
        self.catalog[view_name] = cls
        return cls

    # -----------------------------------------------------------------------
    # Create a Get-Table from YAML definition
    # -----------------------------------------------------------------------

    def _build_optable(self, table_name):
        """build a new Get-Table definition"""
        if table_name in self.catalog:
            return self.catalog[table_name]

        tbl_dict = self._catalog_dict[table_name]
        kvargs = deepcopy(tbl_dict)

        rpc = kvargs.pop("rpc")
        kvargs["table_name"] = table_name

        if "view" in tbl_dict:
            view_name = tbl_dict["view"]
            cls_view = self.catalog.get(view_name, self._build_view(view_name))
            kvargs["view"] = cls_view

        cls = _GET(rpc, **kvargs)
        self.catalog[table_name] = cls
        return cls

    # -----------------------------------------------------------------------
    # Create a Get-Table from YAML definition
    # -----------------------------------------------------------------------

    def _build_cmdtable(self, table_name):
        """build a new command-Table definition"""
        if table_name in self.catalog:
            return self.catalog[table_name]

        tbl_dict = self._catalog_dict[table_name]
        kvargs = deepcopy(tbl_dict)

        if "command" in kvargs:
            cmd = kvargs.pop("command")
            kvargs["table_name"] = table_name

            if "view" in tbl_dict:
                view_name = tbl_dict["view"]
                cls_view = self.catalog.get(view_name, self._build_cmdview(view_name))
                kvargs["view"] = cls_view

            cls = _CMDTBL(cmd, **kvargs)
            self.catalog[table_name] = cls
            return cls
        elif "title" in kvargs:
            cmd = kvargs.pop("title")
            kvargs["table_name"] = table_name

            if "view" in tbl_dict:
                view_name = tbl_dict["view"]
                cls_view = self.catalog.get(view_name, self._build_cmdview(view_name))
                kvargs["view"] = cls_view

            cls = _CMDCHILDTBL(cmd, **kvargs)
            self.catalog[table_name] = cls
            return cls
        else:
            kvargs["table_name"] = table_name

            if "view" in tbl_dict:
                view_name = tbl_dict["view"]
                cls_view = self.catalog.get(view_name, self._build_cmdview(view_name))
                kvargs["view"] = cls_view

            cls = _CMDCHILDTBL(**kvargs)
            self.catalog[table_name] = cls
            return cls

    # -----------------------------------------------------------------------
    # Create a Table class from YAML definition
    # -----------------------------------------------------------------------

    def _build_table(self, table_name):
        """build a new Table definition"""
        if table_name in self.catalog:
            return self.catalog[table_name]

        tbl_dict = self._catalog_dict[table_name]

        table_item = tbl_dict.pop("item")
        kvargs = deepcopy(tbl_dict)
        kvargs["table_name"] = table_name

        if "view" in tbl_dict:
            view_name = tbl_dict["view"]
            cls_view = self.catalog.get(view_name, self._build_view(view_name))
            kvargs["view"] = cls_view

        cls = _TABLE(table_item, **kvargs)
        self.catalog[table_name] = cls
        return cls

    def _build_cfgtable(self, table_name):
        """build a new Config-Table definition"""
        if table_name in self.catalog:
            return self.catalog[table_name]
        tbl_dict = deepcopy(self._catalog_dict[table_name])

        if "view" in tbl_dict:
            # transpose name to class
            view_name = tbl_dict["view"]
            tbl_dict["view"] = self.catalog.get(view_name, self._build_view(view_name))

        cls = _CFGTBL(table_name, tbl_dict)
        self.catalog[table_name] = cls
        return cls

    # -----------------------------------------------------------------------
    # Primary builders ...
    # -----------------------------------------------------------------------

    def _sortitems(self):
        for k, v in self._catalog_dict.items():
            if "rpc" in v:
                self._item_optables.append(k)
            elif "get" in v:
                self._item_cfgtables.append(k)
            elif "set" in v:
                self._item_cfgtables.append(k)
            elif "command" in v or "title" in v:
                self._item_cmdtables.append(k)
            elif "view" in v and "item" in v and v["item"] == "*":
                self._item_cmdtables.append(k)
            elif "view" in v:
                self._item_tables.append(k)
            else:
                self._item_views.append(k)

    def load(self, catalog_dict, envrion={}):
        # load the yaml data and extract the item names.  these names will
        # become the new class definitions

        self._catalog_dict = catalog_dict
        self._sortitems()

        list(map(self._build_optable, self._item_optables))
        list(map(self._build_cfgtable, self._item_cfgtables))
        list(map(self._build_cmdtable, self._item_cmdtables))
        list(map(self._build_table, self._item_tables))
        list(map(self._build_view, self._item_views))

        return self.catalog



from jnpr.junos.device import Device
from jnpr.junos.console import Console
from jnpr.junos.factory.to_json import PyEzJSONEncoder
from jnpr.junos.facts.swver import version_info
from jnpr.junos.facts.swver import version_yaml_representer
from . import jxml
from . import jxml as JXML
from . import version
from . import exception

import json
import yaml
import logging

__date__ = version.DATE

# import time
# __date__ = time.strftime("%Y-%b-%d")

# Set default JSON encoder
json._default_encoder = PyEzJSONEncoder()

# Disable ignore_aliases for YAML dumper
# To support version_info
yaml.dumper.SafeDumper.ignore_aliases = lambda self, data: True
yaml.dumper.Dumper.ignore_aliases = lambda self, data: True
# Add YAML representer for version_info
yaml.Dumper.add_multi_representer(version_info, version_yaml_representer)
yaml.SafeDumper.add_multi_representer(version_info, version_yaml_representer)


# Suppress Paramiko logger warnings
plog = logging.getLogger("paramiko")
if not plog.handlers:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

    plog.addHandler(NullHandler())

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



"""
A dictionary-like object of read-only facts about the Junos device.

These facts are accessed as the `facts` attribute of a `Device` object
instance. For example, if `dev` is an instance of a `Device` object,
the hostname of the device can be accessed with::

    dev.facts['hostname']

Force a refresh of all facts with::

    dev.facts_refresh()

Force a refresh of a single fact with::

    dev.facts_refresh(keys='hostname')

Force a refresh of a set of facts with::

    dev.facts_refresh(keys=('hostname','domain','fqdn'))

NOTE: The dictionary key for each available fact is guaranteed to exist. If
      there is a problem gathering the value of a specific fact/key, or if
      the fact is not supported on a given platform, then the fact/key will
      have the value None (the None object, not a string.)

      Accessing a dictionary key which does not correspond to an available fact
      will raise a KeyError (the same behavior as accessing a non-existent key
      of a normal dict.)

The following dictionary keys represent the available facts and their meaning:

"""
import sys

import jnpr.junos.facts.current_re
import jnpr.junos.facts.domain
import jnpr.junos.facts.ethernet_mac_table
import jnpr.junos.facts.file_list
import jnpr.junos.facts.get_chassis_cluster_status
import jnpr.junos.facts.get_chassis_inventory
import jnpr.junos.facts.get_route_engine_information
import jnpr.junos.facts.get_software_information
import jnpr.junos.facts.get_virtual_chassis_information
import jnpr.junos.facts.ifd_style
import jnpr.junos.facts.iri_mapping
import jnpr.junos.facts.personality
import jnpr.junos.facts.swver
import jnpr.junos.facts.is_linux


def _build_fact_callbacks_and_doc_strings():
    """
    Imports the fact modules and returns callbacks and doc_strings.

    :returns:
      A tuple of callbacks and doc_strings.
      callbacks - a dict of the callback function to invoke for each fact.
      doc_strings - a dict of the doc string for each fact.

    :raises:
      RuntimeError if more than one module claims to provide the same fact.
                   This is an indication of incorrectly written fact module(s).
                   In order to remain deterministic, each fact must be
                   provided by a single module.
    """
    callbacks = {}
    doc_strings = {}
    for name, module in sys.modules.items():
        if name.startswith("jnpr.junos.facts.") and module is not None:
            new_doc_strings = module.provides_facts()
            for key in new_doc_strings:
                if key not in callbacks:
                    callbacks[key] = module.get_facts
                    doc_strings[key] = new_doc_strings[key]
                else:
                    raise RuntimeError(
                        "Both the %s module and the %s module "
                        "claim to provide the %s fact. Please "
                        "report this error."
                        % (callbacks[key].__module__, module.__name__, key)
                    )
    return (callbacks, doc_strings)


# Import all of the fact modules and build the callbacks and doc strings
(_callbacks, _doc_strings) = _build_fact_callbacks_and_doc_strings()

# In case optimization flag is enabled, it strips of docstring and __doc__ becomes None
if __doc__ is None:
    __doc__ = ""

# Append the doc string (__doc__) with the documentation for each fact.
for key in sorted(_doc_strings, key=lambda s: s.lower()):
    __doc__ += ":%s:\n  %s\n" % (key, _doc_strings[key])



"""
  facts['HOME'] = login home directory
"""
from lxml.builder import E


def facts_session(dev, facts):
    facts["HOME"] = dev.rpc(E.command("show cli directory")).findtext(
        "./working-directory"
    )



import re
import logging

from jnpr.junos.facts.swver import version_info
from jnpr.junos.exception import RpcError
from jnpr.junos import jxml as JXML

logger = logging.getLogger("jnpr.junos.facts.get_software_information")


def _get_software_information(device):
    # See if device understands "invoke-on all-routing-engines"
    try:
        return device.rpc.cli(
            "show version invoke-on all-routing-engines", format="xml", normalize=True
        )
    except RpcError as err:
        # See if device runs on a linux kernel
        if device.facts["_is_linux"]:
            sw_info_all = device.rpc.get_software_information(
                normalize=True, node="all"
            )
            sw_info_re0 = device.rpc.get_software_information(
                normalize=True, node="re0"
            )
            sw_info_re1 = device.rpc.get_software_information(
                normalize=True, node="re1"
            )
            re0_hostname = sw_info_re0.findtext("./host-name")
            re1_hostname = sw_info_re1.findtext("./host-name")
            for current_hostname in sw_info_all.findall(
                "./multi-routing-engine-result/software-information/host-name"
            ):
                if current_hostname.text == re0_hostname:
                    current_hostname.getparent().getparent().append(
                        JXML("<re-name>re0</re-name>")
                    )
                elif current_hostname.text == re1_hostname:
                    current_hostname.getparent().getparent().append(
                        JXML("<re-name>re1</re-name>")
                    )
            return sw_info_all
        # See if device is VC Capable
        if device.facts["vc_capable"] is True:
            try:
                return device.rpc.cli(
                    "show version all-members", format="xml", normalize=True
                )
            except Exception:
                pass
        # check if rpc-reply got 2 child element, one rpc-error and another
        # software information
        elif (
            hasattr(err, "rpc_error")
            and err.rpc_error is not None
            and "Could not connect to " in err.rpc_error.get("message")
        ):
            logger.debug(err.rpc_error.get("message"))
            # getparent as rpc-reply got software-information in 2nd element
            # and dev.cli return just 1st element.
            rsp = err.xml.getparent()
            rsp = JXML.remove_namespaces(rsp)
            if rsp.xpath(".//software-information"):
                return rsp
        try:
            # JDM for Junos Node Slicing
            return device.rpc.get_software_information(
                all_servers=True, format="xml", normalize=True
            )
        except Exception:
            pass
        try:
            sw_info = device.rpc.get_software_information(normalize=True)
        except Exception:
            sw_info = True
        try:
            if sw_info is True:
                # Possibly an NFX which requires 'local' and 'detail' args.
                sw_info = device.rpc.get_software_information(
                    local=True, detail=True, normalize=True
                )
            return sw_info
        except Exception:
            pass


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "junos_info": "A two-level dictionary providing Junos software "
        "version information for each RE in the system. "
        "The first-level key is the name of the RE. The "
        "second level key is 'text' for the version as a "
        "string and 'object' for the version as a "
        "version_info object.",
        "hostname": "A string containing the hostname of the current "
        "Routing Engine.",
        "hostname_info": "A dictionary keyed on Routing Engine name. The "
        "value of each key is the hostname of the "
        "Routing Engine.",
        "model": "An uppercase string containing the model of the chassis "
        "in which the current Routing Engine resides.",
        "model_info": "A dictionary keyed on Routing Engine name. The "
        "value of each key is an uppercase string "
        "containing the model of the chassis in which the "
        "Routing Engine resides.",
        "version": "A string containing the Junos version of the current "
        "Routing Engine.",
        "version_info": "The Junos version of the current Routing Engine "
        "as a version_info object.",
        "version_RE0": "A string containing the Junos version of the "
        "RE in slot 0. (Assuming the system contains an "
        "RE0.)",
        "version_RE1": "A string containing the Junos version of the "
        "RE in slot 1. (Assuming the system contains an "
        "RE1)",
    }


def get_facts(device):
    """
    Gathers facts from the <get-software-information/> RPC.
    """
    junos_info = None
    hostname = None
    hostname_info = None
    model = None
    model_info = None
    version = None
    ver_info = None
    version_RE0 = None
    version_RE1 = None

    rsp = _get_software_information(device)

    if rsp.tag == "software-information":
        si_rsp = [rsp]
    else:
        si_rsp = rsp.findall(".//software-information")

    for re_sw_info in si_rsp:
        re_name = re_sw_info.findtext("../re-name", "re0")
        re_model = re_sw_info.findtext("./product-model")
        re_hostname = re_sw_info.findtext("./host-name")
        # First try the <junos-version> tag present in >= 15.1
        re_version = re_sw_info.findtext("./junos-version")
        if re_version is None:
            # For < 15.1, get version from the "junos" package.
            try:
                re_pkg_info = re_sw_info.findtext(
                    './package-information[name="junos"]/comment'
                )
                if re_pkg_info is not None:
                    re_version = re.findall(r"\[(.*)\]", re_pkg_info)[0]
                else:
                    # Junos Node Slicing JDM case
                    re_pkg_info = re_sw_info.findtext(
                        './package-information[name="JUNOS version"]/comment'
                    )
                    if re_pkg_info is not None:
                        # In this case, re_pkg_info might look like this:
                        # JUNOS version : 17.4-20170703_dev_common.0-secure
                        # Match everything from last space until the end.
                        re_version = re.findall(r".*\s+(.*)", re_pkg_info)[0]
                    else:
                        # NFX JDM case
                        re_pkg_info = re_sw_info.findtext(
                            './version-information[component="MGD"]/release'
                        )
                        re_version = re.findall(r"(.*\d+)", re_pkg_info)[0]
            except Exception:
                re_version = None
        if model_info is None and re_model is not None:
            model_info = {}
        if re_model is not None:
            model_info[re_name] = re_model.upper()
        if hostname_info is None and re_hostname is not None:
            hostname_info = {}
        if re_hostname is not None:
            hostname_info[re_name] = re_hostname
        if junos_info is None and re_version is not None:
            junos_info = {}
        if re_version is not None:
            junos_info[re_name] = {
                "text": re_version,
                "object": version_info(re_version),
            }

        # Check to see if re_name is the RE we are currently connected to.
        # There are at least five cases to handle.
        this_re = False
        # 1) this device doesn't support the current_re fact and there's only
        #    one RE.
        if device.facts["current_re"] is None and len(si_rsp) == 1:
            this_re = True
        # 2) re_name is in the current_re fact. The easy case.
        elif re_name in device.facts["current_re"]:
            this_re = True
        # 3) Some single-RE devices (discovered on EX2200 running 11.4R1)
        # don't include 'reX' in the current_re list. Check for this
        # condition and still set default hostname, model, and version
        elif (
            re_name == "re0"
            and "re1" not in device.facts["current_re"]
            and "master" in device.facts["current_re"]
        ):
            this_re = True
        # 4) For an lcc in a TX(P) re_name is 're0' or 're1', but the
        # current_re fact is ['lcc1-re0', 'member1-re0', ...]. Check to see
        # if any iri_name endswith the re_name.
        if this_re is False:
            for iri_name in device.facts["current_re"]:
                if iri_name.endswith("-" + re_name):
                    this_re = True
                    break
        # 5) For an MX configured with Node Virtualization then, re_name is
        #    bsys-reX, but the iri_name is still just reX.
        if this_re is False:
            for iri_name in device.facts["current_re"]:
                if re_name == "bsys-" + iri_name:
                    this_re = True
                    break
        # Set hostname, model, and version facts if we've found the RE to
        # which we are currently connected.
        if this_re is True:
            if hostname is None:
                hostname = re_hostname
            if model is None:
                model = re_model.upper()
            if version is None:
                version = re_version

    if version is None:
        version = "0.0I0.0"
    ver_info = version_info(version)
    if junos_info is not None:
        if "re0" in junos_info:
            version_RE0 = junos_info["re0"]["text"]
        elif "node0" in junos_info:
            version_RE0 = junos_info["node0"]["text"]
        elif "bsys-re0" in junos_info:
            if len(junos_info) > 4:
                version_RE0 = junos_info["bsys-re0"]["text"]
            else:
                for key in junos_info.keys():
                    if key.startswith("gnf") and key.endswith("re0"):
                        version_RE0 = junos_info[key]["text"]
        elif "server0" in junos_info:
            version_RE0 = junos_info["server0"]["text"]
        if "re1" in junos_info:
            version_RE1 = junos_info["re1"]["text"]
        elif "node1" in junos_info:
            version_RE1 = junos_info["node1"]["text"]
        elif "bsys-re1" in junos_info:
            if len(junos_info) > 4:
                version_RE1 = junos_info["bsys-re1"]["text"]
            else:
                for key in junos_info.keys():
                    if key.startswith("gnf") and key.endswith("re1"):
                        version_RE1 = junos_info[key]["text"]
        elif "server1" in junos_info:
            version_RE1 = junos_info["server1"]["text"]

    return {
        "junos_info": junos_info,
        "hostname": hostname,
        "hostname_info": hostname_info,
        "model": model,
        "model_info": model_info,
        "version": version,
        "version_info": ver_info,
        "version_RE0": version_RE0,
        "version_RE1": version_RE1,
    }



import sys
import os
import yaml
import types

from jnpr.junos.factory.factory_loader import FactoryLoader

__all__ = []


class MetaPathFinder(object):
    def find_module(self, fullname, path=None):
        if fullname.startswith("jnpr.junos"):
            mod = fullname.split(".")[-1]
            if mod in [
                os.path.splitext(i)[0] for i in os.listdir(os.path.dirname(__file__))
            ]:
                return MetaPathLoader()


class MetaPathLoader(object):
    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]
        mod = fullname.split(".")[-1]
        modObj = types.ModuleType(
            mod, "Module created to provide a context for %s" % mod
        )
        with open(os.path.join(os.path.dirname(__file__), mod + ".yml"), "r") as stream:
            try:
                modules = FactoryLoader().load(
                    yaml.load(stream, Loader=yaml.FullLoader)
                )
            except yaml.YAMLError as exc:
                raise ImportError("%s is not loaded" % mod)
        for k, v in modules.items():
            setattr(modObj, k, v)
        sys.modules[fullname] = modObj
        return modObj


sys.meta_path.insert(0, MetaPathFinder())



from jnpr.junos.utils.fs import FS
from jnpr.junos.exception import RpcError
from jnpr.junos.jxml import INHERIT
from lxml.builder import E


def facts_domain(junos, facts):
    """
    The following facts are required:
        facts['hostname']

    The following facts are assigned:
        facts['domain']
        facts['fqdn']
    """

    try:
        domain_filter_xml = E("configuration", E("system", E("domain-name")))
        domain = junos.rpc.get_config(filter_xml=domain_filter_xml, options=INHERIT)
        domain_name = domain.xpath(".//domain-name")
        if len(domain_name) > 0:
            facts["domain"] = domain_name[0].text
            facts["fqdn"] = facts["hostname"] + "." + facts["domain"]
            return
    except RpcError:
        pass

    fs = FS(junos)
    file_content = fs.cat("/etc/resolv.conf") or fs.cat("/var/etc/resolv.conf")
    words = file_content.split() if file_content is not None else ""
    if "domain" not in words:
        facts["domain"] = None
        facts["fqdn"] = facts["hostname"]
    else:
        idx = words.index("domain") + 1
        facts["domain"] = words[idx]
        facts["fqdn"] = facts["hostname"] + "." + facts["domain"]



def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "_iri_hostname": "A dictionary keyed by internal routing instance "
        "ip addresses. The value of each key is the "
        "internal routing instance hostname for the ip",
        "_iri_ip": "A dictionary keyed by internal routing instance "
        "hostnames. The value of each key is the internal "
        "routing instance ip for the hostname",
    }


def get_facts(device):
    """
    Gathers facts from a <file-show/> RPC on the '/etc/hosts.junos' file.
    """
    iri_hostname = None
    iri_ip = None

    rsp = device.rpc.file_show(filename="/etc/hosts.junos", normalize=False)

    if rsp is not None:
        hosts_file_content = rsp.findtext(".", default="")
        if hosts_file_content is not None:
            for line in hosts_file_content.splitlines():
                (line, _, _) = line.partition("#")
                components = line.split(None)
                if len(components) > 1:
                    ip = components[0]
                    hosts = components[1:]
                    if iri_hostname is None:
                        iri_hostname = {}
                    if iri_ip is None:
                        iri_ip = {}
                    if ip in iri_hostname:
                        iri_hostname[ip] += hosts
                    else:
                        iri_hostname[ip] = hosts
                    for host in hosts:
                        if host in iri_ip:
                            iri_ip[host].append(ip)
                        else:
                            iri_ip[host] = [ip]
                    for host in hosts:
                        # Handle templates with %d
                        if "%d" in host:
                            octets = ip.split(".", 3)
                            for count in range(255):
                                t_ip = (
                                    octets[0]
                                    + "."
                                    + octets[1]
                                    + "."
                                    + str(count)
                                    + "."
                                    + octets[3]
                                )
                                t_host = host.replace("%d", str(count))
                                if t_ip in iri_hostname:
                                    iri_hostname[t_ip].append(t_host)
                                else:
                                    iri_hostname[t_ip] = [t_host]
                                if t_host in iri_ip:
                                    iri_ip[t_host].append(t_ip)
                                else:
                                    iri_ip[t_host] = [t_ip]

    return {
        "_iri_hostname": iri_hostname,
        "_iri_ip": iri_ip,
    }



import re


def facts_switch_style(junos, facts):
    persona = facts["personality"]

    if persona in ["MX", "SRX_HIGHEND"]:
        style = "BRIDGE_DOMAIN"
    elif persona in ["SWITCH", "SRX_BRANCH"]:
        model = facts["model"]
        if re.match("firefly", model, re.IGNORECASE):
            style = "NONE"
        elif re.match("^(EX9)|(EX43)", model):
            style = "VLAN_L2NG"
        else:
            style = "VLAN"
    else:
        style = "NONE"

    facts["switch_style"] = style



from lxml.builder import E

from jnpr.junos.utils.util import Util
from jnpr.junos.utils.start_shell import StartShell

from jnpr.junos.exception import RpcError


class FS(Util):

    """
    Filesystem (FS) utilities:

    * :meth:`cat`: show the contents of a file
    * :meth:`checksum`: calculate file checksum (md5,sha256,sha1)
    * :meth:`cp`: local file copy (not scp)
    * :meth:`cwd`: change working directory
    * :meth:`ls`: return file/dir listing
    * :meth:`mkdir`: create a directory
    * :meth:`pwd`: get working directory
    * :meth:`mv`: local file rename
    * :meth:`rm`: local file delete
    * :meth:`rmdir`: remove a directory
    * :meth:`stat`: return file/dir information
    * :meth:`storage_usage`: return storage usage
    * :meth:`directory_usage`: return directory usage
    * :meth:`storage_cleanup`: perform storage storage_cleanup
    * :meth:`storage_cleanup_check`: returns a list of files which will be
                                     removed at cleanup
    * :meth:`symlink`: create a symlink
    * :meth:`tgz`: tar+gzip a directory

    .. note: The following methods require 'start shell' priveldges:

              * mkdir
              * rmdir
              * symlink
    """

    # -------------------------------------------------------------------------
    # cat - show file contents
    # -------------------------------------------------------------------------

    def cat(self, path):
        """
        Returns the contents of the file **path**.

        :param str path: File-path

        :returns: contents of the file (str) or ``None`` if file does not exist
        """
        try:
            rsp = self._dev.rpc.file_show(filename=path)
        except:
            return None
        return rsp.text

    # -------------------------------------------------------------------------
    # cwd - change working directory
    # -------------------------------------------------------------------------

    def cwd(self, path):
        """
        Change working directory to **path**.

        :param str path: path to working directory
        """
        rsp = self._dev.rpc.set_cli_working_directory(directory=path)
        return rsp.findtext("working-directory")

    # -------------------------------------------------------------------------
    # pwd - return current working directory
    # -------------------------------------------------------------------------

    def pwd(self):
        """
        :returns: The current working directory path (str)
        """
        rsp = self._dev.rpc(E.command("show cli directory"))
        return rsp.findtext("./working-directory")

    # -------------------------------------------------------------------------
    # checksum - compute file checksum
    # -------------------------------------------------------------------------

    def checksum(self, path, calc="md5"):
        """
        Performs the checksum command on the given file path using the
        required calculation method and returns the string value.
        If the **path** is not found on the device, then ``None`` is returned.

        :param str path: file-path on local device
        :param str calc: checksum calculation method:

                         * "md5"
                         * "sha256"
                         * "sha1"

        :returns: checksum value (str) or ``None`` if file not found
        """
        cmd_map = {
            "md5": self._dev.rpc.get_checksum_information,
            "sha256": self._dev.rpc.get_sha256_checksum_information,
            "sha1": self._dev.rpc.get_sha1_checksum_information,
        }
        rpc = cmd_map.get(calc)
        if rpc is None:
            raise ValueError("Unknown calculation method: '%s'" % calc)
        try:
            rsp = rpc(path=path)
            return rsp.findtext(".//checksum").strip()
        except:
            # the only exception is that the path is not found
            return None

    @classmethod
    def _decode_file(cls, fileinfo):
        results = {}

        not_file = fileinfo.xpath("file-directory | file-symlink-target")
        if len(not_file):
            results["type"] = {"file-directory": "dir", "file-symlink-target": "link"}[
                not_file[0].tag
            ]
            if "link" == results["type"]:
                results["link"] = not_file[0].text.strip()
        else:
            results["type"] = "file"

        results["path"] = fileinfo.findtext("file-name").strip()
        results["owner"] = fileinfo.findtext("file-owner").strip()
        results["size"] = int(fileinfo.findtext("file-size"))
        fper = fileinfo.find("file-permissions")
        results["permissions"] = int(fper.text.strip())
        results["permissions_text"] = fper.get("format")
        fdate = fileinfo.find("file-date")
        results["ts_date"] = fdate.get("format")
        results["ts_epoc"] = fdate.text.strip()
        return results

    @classmethod
    def _decode_dir(cls, dirinfo, files=None):
        results = {}
        results["type"] = "dir"
        results["path"] = dirinfo.get("name")
        if files is None:
            files = dirinfo.xpath("file-information")
        results["file_count"] = len(files)
        results["size"] = sum([int(f.findtext("file-size")) for f in files])
        return results

    # -------------------------------------------------------------------------
    # stat - file information
    # -------------------------------------------------------------------------

    def stat(self, path):
        """
        Returns a dictionary of status information on the path, or ``None``
        if the path does not exist.

        :param str path: file-path on local device

        :returns: status information on the file
        :rtype: dict
        """
        rsp = self._dev.rpc.file_list(detail=True, path=path)

        # if there is an output tag, then it means that the path
        # was not found
        if rsp.find("output") is not None:
            return None

        # ok, so we've either got a directory or a file at
        # this point, so decode accordingly

        xdir = rsp.find("directory")
        if xdir.get("name"):  # then this is a directory path
            return FS._decode_dir(xdir)
        else:
            return FS._decode_file(xdir.find("file-information"))

    # -------------------------------------------------------------------------
    # ls - file/dir listing
    # -------------------------------------------------------------------------

    def ls(self, path=".", brief=False, followlink=True):
        """
        File listing, returns a dict of file information.  If the
        path is a symlink, then by default **followlink** will
        recursively call this method to obtain the symlink specific
        information.

        :param str path:
            file-path on local device. defaults to current
            working directory
        :param bool brief:
            when ``True`` brief amount of data
        :param bool followlink:
            when ``True`` (default) this method will recursively
            follow the directory symlinks to gather data

        :returns: dict collection of file information or ``None``
                  if **path** is not found
        """
        rsp = self._dev.rpc.file_list(detail=True, path=path)

        # if there is an output tag, then it means that the path
        # was not found, and we return :None:

        if rsp.find("output") is not None:
            return None

        xdir = rsp.find(".//directory")

        # check to see if the directory element has a :name:
        # attribute, and if it does not, then this is a file, and
        # decode accordingly.  If the file is a symlink, then we
        # want to follow the symlink to get what we want.

        if not xdir.get("name"):
            results = FS._decode_file(xdir.find("file-information"))
            link_path = results.get("link")
            if not link_path:  # then we are done
                return results
            else:
                return results if followlink is False else self.ls(path=link_path)

        # if we are here, then it's a directory, include information on all
        # files
        files = xdir.xpath("file-information")
        results = FS._decode_dir(xdir, files)

        if brief is True:
            results["files"] = [f.findtext("file-name").strip() for f in files]
        else:
            results["files"] = dict(
                (f.findtext("file-name").strip(), FS._decode_file(f)) for f in files
            )

        return results

    # -------------------------------------------------------------------------
    # storage_usage - filesystem storage usage
    # -------------------------------------------------------------------------

    def storage_usage(self):
        """
        Returns the storage usage, similar to the unix "df" command.

        :returns: dict of storage usage
        """
        rsp = self._dev.rpc.get_system_storage()

        def _name(fs):
            return fs.findtext("filesystem-name").strip()

        def _decode(fs):
            r = {}
            r["mount"] = fs.find("mounted-on").text.strip()
            tb = fs.find("total-blocks")
            r["total"] = tb.get("format")
            r["total_blocks"] = int(tb.text)
            ub = fs.find("used-blocks")
            r["used"] = ub.get("format")
            r["used_blocks"] = int(ub.text)
            r["used_pct"] = fs.find("used-percent").text.strip()
            ab = fs.find("available-blocks")
            r["avail"] = ab.get("format")
            r["avail_block"] = int(ab.text)
            return r

        re_list = rsp.xpath("multi-routing-engine-item")
        if re_list:
            fs_dict = {}
            for re in re_list:
                re_name = re.findtext("re-name").strip()
                re_fs_dict = dict(
                    (_name(fs), _decode(fs))
                    for fs in re.xpath("system-storage-information/filesystem")
                )
                fs_dict[re_name] = re_fs_dict
            return fs_dict

        return dict((_name(fs), _decode(fs)) for fs in rsp.xpath("filesystem"))

    # -------------------------------------------------------------------------
    # directory_usage - filesystem directory usage
    # -------------------------------------------------------------------------

    def directory_usage(self, path=".", depth=0):
        """
        Returns the directory usage, similar to the unix "du" command.

        :returns: dict of directory usage, including subdirectories
                  if depth > 0
        """
        BLOCK_SIZE = 512

        rsp = self._dev.rpc.get_directory_usage_information(path=path, depth=str(depth))

        result = {}

        directories = rsp.findall(".//directory")
        if not directories:
            raise RpcError(rsp=rsp)

        for directory in directories:
            dir_name = directory.findtext("directory-name")
            if dir_name is not None:
                dir_name = dir_name.strip()
            else:
                raise RpcError(rsp=rsp)

            used_space = directory.find("used-space")
            if used_space is not None:
                dir_size = used_space.text.strip()
                dir_blocks = used_space.get("used-blocks")
                if dir_blocks is not None:
                    dir_blocks = int(dir_blocks)
                    dir_bytes = dir_blocks * BLOCK_SIZE
                    result[dir_name.strip()] = {
                        "size": dir_size,
                        "blocks": dir_blocks,
                        "bytes": dir_bytes,
                    }

        return result

    # -------------------------------------------------------------------------
    # storage_cleanup_check, storage_cleanip
    # -------------------------------------------------------------------------

    @classmethod
    def _decode_storage_cleanup(cls, files):
        def _name(f):
            return f.findtext("file-name").strip()

        def _decode(f):
            return {
                "size": int(f.findtext("size")),
                "ts_date": f.findtext("date").strip(),
            }

        # return a dict of name/decode pairs for each file
        return dict((_name(f), _decode(f)) for f in files)

    def storage_cleanup_check(self):
        """
        Perform the 'request system storage cleanup dry-run' command
        to return a ``dict`` of files/info that would be removed if
        the cleanup command was executed.

        :returns: dict of files that would be removed (dry-run)
        """
        rsp = self._dev.rpc.request_system_storage_cleanup(dry_run=True)
        files = rsp.xpath("file-list/file")
        return FS._decode_storage_cleanup(files)

    def storage_cleanup(self):
        """
        Perform the 'request system storage cleanup' command to remove
        files from the filesystem.  Return a ``dict`` of file name/info
        on the files that were removed.

        :returns: dict on files that were removed
        """
        rsp = self._dev.rpc.request_system_storage_cleanup()
        files = rsp.xpath("file-list/file")
        return FS._decode_storage_cleanup(files)

    # -------------------------------------------------------------------------
    # rm - local file delete
    # -------------------------------------------------------------------------

    def rm(self, path):
        """
        Performs a local file delete action, per Junos CLI command
        "file delete".

        :returns: ``True`` when successful, ``False`` otherwise.
        """
        # the return value from this RPC will return either True if the delete
        # was successful, or an XML structure otherwise.  So we can do a simple
        # test to provide the return result to the caller.
        rsp = self._dev.rpc.file_delete(path=path)
        if rsp is True:
            return True
        else:
            return False

    # -------------------------------------------------------------------------
    # cp - local file copy
    # -------------------------------------------------------------------------

    def cp(self, from_path, to_path):
        """
        Perform a local file copy where **from_path** and **to_path** can be
        any valid Junos path argument.  Refer to the Junos "file copy" command
        documentation for details.

        :param str from_path: source file-path
        :param str to_path: destination file-path

        .. notes: Valid Junos file-path can include URL, such as ``http://``.
                  this is handy for copying files for webservers.

        :returns: ``True`` if OK, ``False`` if file does not exist.
        """
        # this RPC returns True if it is OK.  If the file does not exist
        # this RPC will generate an RpcError exception, so just return False
        try:
            self._dev.rpc.file_copy(source=from_path, destination=to_path)
        except:
            return False
        return True

    # -------------------------------------------------------------------------
    # mv - local file rename
    # -------------------------------------------------------------------------

    def mv(self, from_path, to_path):
        """
        Perform a local file rename function, same as "file rename" Junos CLI.

        :returns: ``True`` if OK, ``False`` if file does not exist.
        """
        rsp = self._dev.rpc.file_rename(source=from_path, destination=to_path)
        if rsp is True:
            return True
        else:
            return False

    def tgz(self, from_path, tgz_path):
        """
        Create a file called **tgz_path** that is the tar-gzip of the given
        directory specified **from_path**.

        :param str from_path: file-path to directory of files
        :param str tgz_path: file-path name of tgz file to create

        :returns: ``True`` if OK, error-msg (str) otherwise
        """
        rsp = self._dev.rpc.file_archive(
            compress=True, source=from_path, destination=tgz_path
        )

        # if the rsp is True, then the command executed OK.
        if rsp is True:
            return True

        # otherwise, return the error string to the caller
        return rsp.text

    # -------------------------------------------------------------------------
    # !!!!! methods that use SSH shell commands, require that the user
    # !!!!! has 'start shell' privileges
    # -------------------------------------------------------------------------

    def _ssh_exec(self, command):
        with StartShell(self._dev) as sh:
            got = sh.run(command)
        return got

    def rmdir(self, path):
        """
        Executes the 'rmdir' command on **path**.

        .. warning:: REQUIRES SHELL PRIVILEGES

        :param str path: file-path to directory

        :returns: ``True`` if OK, error-message (str) otherwise
        """
        results = self._ssh_exec("rmdir %s" % path)
        return True if results[0] is True else "".join(results[1][2:-1])

    def mkdir(self, path):
        """
        Executes the 'mkdir -p' command on **path**.

        .. warning:: REQUIRES SHELL PRIVILEGES

        :returns: ``True`` if OK, error-message (str) otherwise
        """
        results = self._ssh_exec("mkdir -p %s" % path)
        return True if results[0] is True else "".join(results[1][2:-1])

    def symlink(self, from_path, to_path):
        """
        Executes the 'ln -sf **from_path** **to_path**' command.

        .. warning:: REQUIRES SHELL PRIVILEGES

        :returns: ``True`` if OK, or error-message (str) otherwise
        """
        results = self._ssh_exec("ln -sf %s %s" % (from_path, to_path))
        return True if results[0] is True else "".join(results[1][2:-1])



from jnpr.junos.exception import RpcError


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "_is_linux": "A boolean indicating if the device is running linux" "kernel.",
    }


def get_facts(device):
    """
    Gathers the _is_linux fact using the <file-show/> RPC on the EVO version file.
    """

    # Temporary implementation until PR 1245634 is implemented.
    LINUX_VERSION_PATH = "/usr/share/cevo/cevo_version"

    is_linux = None

    try:
        rsp = device.rpc.file_show(normalize=True, filename=LINUX_VERSION_PATH)
        if rsp.tag == "file-content":
            is_linux = True
        else:
            is_linux = False
    except RpcError:
        is_linux = False
    return {
        "_is_linux": is_linux,
    }



import re


def facts_personality(junos, facts):
    model = facts["model"]

    if model != "Virtual Chassis":
        examine = model
    else:
        for fact in facts:
            if re.match(r"^RE\d", fact):
                examine = facts[fact]["model"]
                break

    examine = examine.upper()

    if re.match("^(EX)|(QFX)", examine):
        persona = "SWITCH"
    elif examine.startswith("MX"):
        persona = "MX"
    elif examine.startswith("VMX"):
        facts["virtual"] = True
        persona = "MX"
    elif examine.startswith("VJX"):
        facts["virtual"] = True
        persona = "SRX_BRANCH"
    elif "VRR" == examine:
        persona = "MX"
        facts["virtual"] = True
    elif examine.startswith("M"):
        persona = "M"
    elif examine.startswith("T"):
        persona = "T"
    elif examine.startswith("PTX"):
        persona = "PTX"
    elif re.match(r"SRX\s?(\d){4}", examine):
        persona = "SRX_HIGHEND"
    elif re.match(r"SRX\s?(\d){3}", examine):
        persona = "SRX_BRANCH"
    elif re.search("firefly", examine, re.IGNORECASE):
        facts["virtual"] = True
        persona = "SRX_BRANCH"
    elif "OLIVE" == examine:
        facts["virtual"] = True
        persona = "OLIVE"
    else:
        persona = "UNKNOWN"

    facts["personality"] = persona



VERSION = "2.6.7"
DATE = "2023-Feb-28"

# Augment with the internal version if present
try:
    from ._version import get_versions

    VERSION = get_versions()["version"]
    del get_versions
except ImportError:
    # No internal version
    pass



# utils/config.py
import os
import re
import warnings

# 3rd-party modules
from lxml import etree

# package modules
from jnpr.junos.exception import *
from jnpr.junos import jxml as JXML
from jnpr.junos.utils.util import Util

"""
Configuration Utilities
"""


class Config(Util):
    """
    Overview of Configuration Utilities.

    * :meth:`commit`: commit changes
    * :meth:`commit_check`: perform the commit check operation
    * :meth:`diff`: return the diff string between running and candidate config
    * :meth:`load`: load changes into the candidate config
    * :meth:`lock`: take an exclusive lock on the candidate config
    * :meth:`pdiff`: prints the diff string (debug/helper)
    * :meth:`rescue`: controls "rescue configuration"
    * :meth:`rollback`: perform the load rollback command
    * :meth:`unlock`: release the exclusive lock
    """

    # ------------------------------------------------------------------------
    # commit
    # ------------------------------------------------------------------------
    def commit(self, **kvargs):
        """
        Commit a configuration.

        :param str comment: If provided logs this comment with the commit.
        :param int confirm: If provided activates confirm safeguard with
                            provided value as timeout (minutes).
        :param int timeout: If provided the command will wait for completion
                            using the provided value as timeout (seconds).
                            By default the device timeout is used.
        :param bool sync: On dual control plane systems, requests that
                            the candidate configuration on one control plane be
                            copied to the other control plane, checked for
                            correct syntax, and committed on both Routing
                            Engines.
        :param bool force_sync: On dual control plane systems, forces the
                            candidate configuration on one control plane to
                            be copied to the other control plane.
        :param bool full: When true requires all the daemons to check and
                            evaluate the new configuration.
        :param bool detail: When true return commit detail as XML

        :param ignore_warning: A boolean, string or list of string.
          If the value is True, it will ignore all warnings regardless of the
          warning message. If the value is a string, it will ignore
          warning(s) if the message of each warning matches the string. If
          the value is a list of strings, ignore warning(s) if the message of
          each warning matches at least one of the strings in the list.

          For example::

            cu.commit(ignore_warning=True)
            cu.commit(ignore_warning='Advertisement-interval is '
                                     'less than four times')
            cu.commit(ignore_warning=['Advertisement-interval is '
                                      'less than four times',
                                      'Chassis configuration for network '
                                      'services has been changed.'])

          .. note::
            When the value of ignore_warning is a string, or list of strings,
            the string is actually used as a case-insensitive regular
            expression pattern. If the string contains only alpha-numeric
            characters, as shown in the above examples, this results in a
            case-insensitive substring match. However, any regular expression
            pattern supported by the re library may be used for more
            complicated match conditions.

        :returns:
            * ``True`` when successful
            * Commit detail XML (when detail is True)

        :raises CommitError: When errors detected in candidate configuration.
                             You can use the Exception errs variable
                             to identify the specific problems

        .. warning::
            If the function does not receive a reply prior to the timeout
            a RpcTimeoutError will be raised.  It is possible the commit
            was successful.  Manual verification may be required.
        """
        rpc_args = {}

        # if a comment is provided, then include that in the RPC

        comment = kvargs.get("comment")
        if comment:
            rpc_args["log"] = comment

        # if confirm is provided, then setup the RPC args
        # so that Junos will either use the default confirm
        # timeout (confirm=True) or a specific timeout
        # (confirm=<minutes>)

        confirm = kvargs.get("confirm")
        if confirm:
            rpc_args["confirmed"] = True
            confirm_val = str(confirm)
            if "True" != confirm_val:
                rpc_args["confirm-timeout"] = confirm_val

        # if a timeout is provided, then include that in the RPC

        timeout = kvargs.get("timeout")
        if timeout:
            rpc_args["dev_timeout"] = timeout

        # Check for force_sync and sync
        if kvargs.get("force_sync"):
            rpc_args["synchronize"] = True
            rpc_args["force-synchronize"] = True
        elif kvargs.get("sync"):
            rpc_args["synchronize"] = True

        # Check for full
        if kvargs.get("full"):
            rpc_args["full"] = True

        # Check for ignore_warning
        ignore_warn = kvargs.get("ignore_warning")
        if ignore_warn:
            rpc_args["ignore_warning"] = ignore_warn

        rpc_varg = []
        detail = kvargs.get("detail")
        if detail:
            rpc_varg = [{"detail": "detail"}]

        # dbl-splat the rpc_args since we want to pass key/value to metaexec
        # if there is a commit/check error, this will raise an execption

        try:
            rsp = self.rpc.commit_configuration(*rpc_varg, **rpc_args)
        except RpcTimeoutError:
            raise
        except RpcError as err:  # jnpr.junos exception
            if err.rsp is not None and err.rsp.find("ok") is not None:
                # this means there are warnings, but no errors
                return True
            else:
                raise CommitError(cmd=err.cmd, rsp=err.rsp, errs=err.errs)
        except Exception as err:
            # so the ncclient gives us something I don't want.  I'm going to
            # convert it and re-raise the commit error
            if hasattr(err, "xml") and isinstance(err.xml, etree._Element):
                raise CommitError(rsp=err.xml)
            else:
                raise

        if detail:
            return rsp
        else:
            return True

    # -------------------------------------------------------------------------
    # commit check
    # -------------------------------------------------------------------------

    def commit_check(self, **kvargs):
        """
        Perform a commit check.  If the commit check passes, this function
        will return ``True``.  If the commit-check results in warnings, they
        are reported and available in the Exception errs.

        :param int timeout: If provided the command will wait for completion
                            using the provided value as timeout (seconds).

        :returns: ``True`` if commit-check is successful (no errors)
        :raises CommitError: When errors detected in candidate configuration.
                             You can use the Exception errs variable
                             to identify the specific problems
        :raises RpcError: When underlying ncclient has an error
        """
        rpc_args = {}

        # if a timeout is provided, then include that in the RPC

        timeout = kvargs.get("timeout")
        if timeout:
            rpc_args["dev_timeout"] = timeout

        try:
            self.rpc.commit_configuration(check=True, **rpc_args)
        except RpcTimeoutError:
            raise
        except RpcError as err:  # jnpr.junos exception
            if err.rsp is not None and err.rsp.find("ok") is not None:
                # this means there is a warning, but no errors
                return True
            else:
                raise CommitError(cmd=err.cmd, rsp=err.rsp, errs=err.errs)
        except Exception as err:
            # :err: is from ncclient, so extract the XML data
            # and convert into dictionary
            if hasattr(err, "xml") and isinstance(err.xml, etree._Element):
                return JXML.rpc_error(err.xml)
            else:
                raise

        return True

    # -------------------------------------------------------------------------
    # show | compare rollback <number|0*>
    # -------------------------------------------------------------------------

    def diff(self, rb_id=0, ignore_warning=False, use_fast_diff=False):
        """
        Retrieve a diff (patch-format) report of the candidate config against
        either the current active config, or a different rollback.

        :param int rb_id: rollback id [0..49]
        :param bool ignore_warning: Ignore any rpc-error with severity warning
        :param bool use_fast_diff: equivalent to "show | compare use-fast-diff"

        :returns:
            * ``None`` if there is no difference
            * ascii-text (str) if there is a difference
        """

        if rb_id < 0 or rb_id > 49:
            raise ValueError("Invalid rollback #" + str(rb_id))

        rpc_params = dict(compare="rollback", rollback=str(rb_id), format="text")
        if use_fast_diff:
            if rb_id > 0:
                raise ValueError("use_fast_diff can only be used with rb_id 0")
            rpc_params["use-fast-diff"] = "yes"
        try:
            rsp = self.rpc.get_configuration(
                rpc_params,
                ignore_warning=ignore_warning,
            )
        except RpcTimeoutError:
            raise
        except RpcError as err:
            if (
                err.rpc_error["severity"] == "warning"
                and err.message == "mgd: statement must contain additional "
                "statements"
            ):
                # Fix for Issue #655, JDM 15.1X53-D45 responses with
                # extraneous warning message
                return "Unable to parse diff from response!"
            else:
                raise

        # The output is expected to be an etree, it may have empty <configuration-output> tags
        # Adding a preventive check and displaying warning in case the value is bool.
        # This check is added for PR 1564189 (juniper internal) which has issue for certain junos releases.
        if isinstance(rsp, bool):
            warnings.warn(
                "diff shouldn't return boolean as a rpc-reply",
                RuntimeWarning,
            )
            return None

        diff_txt = rsp.find("configuration-output").text
        return None if diff_txt == "\n" else diff_txt

    def pdiff(self, rb_id=0, ignore_warning=False, use_fast_diff=False):
        """
        Helper method that calls ``print`` on the diff (patch-format) between
        the current candidate and the provided rollback.

        :param int rb_id: the rollback id value [0-49]
        :param bool ignore_warning: Ignore any rpc-error with severity warning
        :param bool use_fast_diff: equivalent to "show | compare use-fast-diff"

        :returns: ``None``
        """
        print(self.diff(rb_id, ignore_warning, use_fast_diff))

    # -------------------------------------------------------------------------
    # helper on loading configs
    # -------------------------------------------------------------------------

    def load(self, *vargs, **kvargs):
        """
        Loads changes into the candidate configuration.  Changes can be
        in the form of strings (text,set,xml, json), XML objects, and files.
        Files can be either static snippets of configuration or Jinja2
        templates.  When using Jinja2 Templates, this method will render
        variables into the templates and then load the resulting change;
        i.e. "template building".

        :param object vargs[0]:
            The content to load.  If the contents is a string, the framework
            will attempt to automatically determine the format.  If it is
            unable to determine the format then you must specify the
            **format** parameter.  If the content is an XML object, then
            this method assumes you've structured it correctly;
            and if not an Exception will be raised.

        :param str path:
            Path to file of configuration on the local server.
            The path extension will be used to determine the format of
            the contents:

            * "conf","text","txt" is curly-text-style
            * "set" - ascii-text, set-style
            * "xml" - ascii-text, XML
            * "json" - ascii-text, json

            .. note:: The format can specifically set using **format**.

        :param str format:
            Determines the format of the contents.
            Supported options - text, set, xml, json

            If not provided, internally application will try to find out the format

        :param bool overwrite:
          Determines if the contents completely replace the existing
          configuration.  Default is ``False``.

          .. note:: This option cannot be used if **format** is "set".

        :param bool merge:
          If set to ``True`` will set the load-config action to merge.
          the default load-config action is 'replace'

        :param bool update:
          If set to ``True`` Compare a complete loaded configuration against
          the candidate configuration. For each hierarchy level or
          configuration object that is different in the two configurations,
          the version in the loaded configuration replaces the version in the
          candidate configuration. When the configuration is later committed,
          only system processes that are affected by the changed configuration
          elements parse the new configuration.

          .. note:: This option cannot be used if **format** is "set".

        :param bool patch:
          If set to ``True`` will set the load-config action to load patch.

        :param str template_path:
          Similar to the **path** parameter, but this indicates that
          the file contents are ``Jinja2`` format and will require
          template-rendering.

          .. note:: This parameter is used in conjunction with
                    **template_vars**. The template filename extension will
                    be used to determine the format-style of the contents,
                    or you can override using **format**.

        :param jinja2.Template template:
          A Jinja2 Template object.  Same description as *template_path*,
          except this option you provide the actual Template, rather than
          a path to the template file.

        :param dict template_vars:
          Used in conjunction with the other template options.  This parameter
          contains a dictionary of variables to render into the template.

        :param ignore_warning: A boolean, string or list of string.
          If the value is True, it will ignore all warnings regardless of the
          warning message. If the value is a string, it will ignore
          warning(s) if the message of each warning matches the string. If
          the value is a list of strings, ignore warning(s) if the message of
          each warning matches at least one of the strings in the list.

          For example::

            cu.load(cnf, ignore_warning=True)
            cu.load(cnf, ignore_warning='statement not found')
            cu.load(cnf, ignore_warning=['statement not found',
                                         'statement has no contents; ignored')

          .. note::
            When the value of ignore_warning is a string, or list of strings,
            the string is actually used as a case-insensitive regular
            expression pattern. If the string contains only alpha-numeric
            characters, as shown in the above examples, this results in a
            case-insensitive substring match. However, any regular expression
            pattern supported by the re library may be used for more
            complicated match conditions.

        :param str url:
          Specify the full pathname of the file that contains the configuration
          data to load. The value can be a local file path, an FTP location, or
          a Hypertext Transfer Protocol (HTTP).
          Refer `Doc page <https://www.juniper.net/documentation/en_US/junos/topics/reference/tag-summary/junos-xml-protocol-load-configuration.html>`_
          for more details.

          For example::

            cu.load(url="/var/home/user/golden.conf")
            cu.load(url="ftp://username@ftp.hostname.net/filename")
            cu.load(url="http://username:password@hostname/path/filename")
            cu.load(url="/var/home/user/golden.conf", overwrite=True)

        :returns:
            RPC-reply as XML object.

        :raises: ConfigLoadError: When errors detected while loading candidate
                                  configuration. You can use the Exception
                                  errs variable  to identify the specific
                                  problems.
        """
        rpc_xattrs = {}
        rpc_xattrs["format"] = "xml"  # default to XML format
        rpc_xattrs["action"] = "replace"  # replace is default action

        rpc_contents = None

        actions = filter(
            lambda item: kvargs.get(item, False),
            ("overwrite", "merge", "update", "patch"),
        )
        if len(list(actions)) >= 2:
            raise ValueError("actions can be only one among %s" % ", ".join(actions))

        # support the ability to completely replace the Junos configuration
        # note: this cannot be used if format='set', per Junos API.

        overwrite = kvargs.get("overwrite", False)
        if overwrite is True:
            rpc_xattrs["action"] = "override"
        if kvargs.get("update") is True:
            rpc_xattrs["action"] = "update"
        elif kvargs.get("merge") is True:
            del rpc_xattrs["action"]
        elif kvargs.get("patch") is True:
            rpc_xattrs["action"] = "patch"

        ignore_warning = kvargs.get("ignore_warning", False)

        # ---------------------------------------------------------------------
        # private helpers ...
        # ---------------------------------------------------------------------

        def _lformat_byext(path):
            """determine the format style from the file extension"""
            ext = os.path.splitext(path)[1]
            if ext == ".xml":
                return "xml"
            if ext in [".conf", ".text", ".txt"]:
                return "text"
            if ext in [".set"]:
                return "set"
            if ext in [".json"]:
                return "json"
            raise ValueError("Unknown file contents from extension: %s" % ext)

        def _lset_format(kvargs, rpc_xattrs):
            """setup the kvargs/rpc_xattrs"""
            # when format is given, setup the xml attrs appropriately
            if kvargs["format"] == "set":
                if overwrite is True or kvargs.get("update") is True:
                    raise ValueError(
                        "conflicting args, cannot use 'set' with '%s'"
                        % ("overwrite" if overwrite is True else "update")
                    )
                rpc_xattrs["action"] = "set"
                kvargs["format"] = "text"
            rpc_xattrs["format"] = kvargs["format"]

        def _lset_fromfile(path):
            """setup the kvargs/rpc_xattrs based on path"""
            if "format" not in kvargs:
                # we use the extension to determine the format
                kvargs["format"] = _lformat_byext(path)
                _lset_format(kvargs, rpc_xattrs)

        def _lset_from_rexp(rpc):
            """setup the kvargs/rpc_xattrs using string regular expression"""
            if re.search(r"^\s*<.*>$", rpc, re.MULTILINE):
                kvargs["format"] = "xml"
            elif re.search(
                r"^\s*(set|delete|rename|insert|activate|deactivate"
                r"|annotate|copy|protect|unprotect)\s",
                rpc,
            ):
                kvargs["format"] = "set"
            elif re.search(r"^[a-z:]*\s*[\w-]+\s+\{", rpc, re.I) and re.search(
                r".*}\s*$", rpc
            ):
                kvargs["format"] = "text"
            elif re.search(r"^\s*\{", rpc) and re.search(r".*}\s*$", rpc):
                kvargs["format"] = "json"

        def try_load(rpc_contents, rpc_xattrs, ignore_warning=False):
            try:
                got = self.rpc.load_config(
                    rpc_contents, ignore_warning=ignore_warning, **rpc_xattrs
                )
            except RpcTimeoutError as err:
                raise err
            except RpcError as err:
                raise ConfigLoadError(cmd=err.cmd, rsp=err.rsp, errs=err.errs)
            # Something unexpected happened - raise it up
            except Exception as err:
                raise

            return got

        # ---------------------------------------------------------------------
        # end-of: private helpers
        # ---------------------------------------------------------------------

        if "format" in kvargs:
            _lset_format(kvargs, rpc_xattrs)

        # ---------------------------------------------------------------------
        # if contents are provided as vargs[0], then process that as XML or str
        # ---------------------------------------------------------------------

        if len(vargs):
            # caller is providing the content directly.
            rpc_contents = vargs[0]
            if isinstance(rpc_contents, str):
                if "format" not in kvargs:
                    _lset_from_rexp(rpc_contents)
                    if "format" in kvargs:
                        _lset_format(kvargs, rpc_xattrs)
                    else:
                        raise RuntimeError(
                            "Not able to resolve the config format "
                            "You must define the format of the contents "
                            "explicitly to the function. Ex: format='set'"
                        )
                if kvargs["format"] == "xml":
                    # covert the XML string into XML structure
                    rpc_contents = etree.XML(rpc_contents)

        # ---------------------------------------------------------------------
        # if path is provided, use the static-config file
        # ---------------------------------------------------------------------

        elif "path" in kvargs:
            # then this is a static-config file.  load that as our rpc_contents
            try:
                # Explicitly request Python 3.x universal newline
                rpc_contents = open(kvargs["path"], "r", newline=None).read()
            except TypeError:
                # Fallback to Python 2.x universal newline
                rpc_contents = open(kvargs["path"], "rU").read()
            _lset_fromfile(kvargs["path"])
            if rpc_xattrs["format"] == "xml":
                # covert the XML string into XML structure
                rpc_contents = etree.XML(rpc_contents)

        # ---------------------------------------------------------------------
        # if template_path is provided, then jinja2 load the template, and
        # render the results.  if template_vars are provided, use those
        # in the render process.
        # ---------------------------------------------------------------------

        elif "template_path" in kvargs:
            path = kvargs["template_path"]
            template = self.dev.Template(path)
            rpc_contents = template.render(kvargs.get("template_vars", {}))
            _lset_fromfile(path)
            if rpc_xattrs["format"] == "xml":
                # covert the XML string into XML structure
                rpc_contents = etree.XML(rpc_contents)

        # ---------------------------------------------------------------------
        # if template is provided, then this is a pre-loaded jinja2 Template
        # object.  Use the template.filename to determine the format style
        # ---------------------------------------------------------------------

        elif "template" in kvargs:
            template = kvargs["template"]
            path = template.filename
            rpc_contents = template.render(kvargs.get("template_vars", {}))
            _lset_fromfile(path)
            if rpc_xattrs["format"] == "xml":
                # covert the XML string into XML structure
                rpc_contents = etree.XML(rpc_contents)

        if rpc_contents is not None:
            return try_load(rpc_contents, rpc_xattrs, ignore_warning=ignore_warning)
        elif "url" in kvargs and rpc_contents is None:
            url = kvargs["url"]
            _lset_fromfile(url)
            rpc_xattrs["url"] = url
            return try_load(rpc_contents, rpc_xattrs, ignore_warning=ignore_warning)
        else:
            raise RuntimeError("Unhandled load request")

    # -------------------------------------------------------------------------
    # config exclusive
    # -------------------------------------------------------------------------

    def lock(self):
        """
        Attempts an exclusive lock on the candidate configuration.  This
        is a non-blocking call.

        :returns:
            ``True`` always when successful

        :raises LockError: When the lock cannot be obtained
        """
        try:
            self.rpc.lock_configuration()
        except Exception as err:
            if isinstance(err, RpcError):
                raise LockError(rsp=err.rsp)
            elif isinstance(err, ConnectClosedError):
                raise err
            else:
                # :err: is from ncclient
                raise LockError(rsp=JXML.remove_namespaces(err.xml))

        return True

    # -------------------------------------------------------------------------
    # releases the exclusive lock
    # -------------------------------------------------------------------------

    def unlock(self):
        """
        Unlocks the candidate configuration.

        :returns:
            ``True`` always when successful

        :raises UnlockError: If you attempt to unlock a configuration
                             when you do not own the lock
        """
        try:
            self.rpc.unlock_configuration()
        except Exception as err:
            if isinstance(err, RpcError):
                raise UnlockError(rsp=err.rsp)
            elif isinstance(err, ConnectClosedError):
                raise err
            else:
                # :err: is from ncclient
                raise UnlockError(rsp=JXML.remove_namespaces(err.xml))

        return True

    # -------------------------------------------------------------------------
    # rollback <number|0*>
    # -------------------------------------------------------------------------

    def rollback(self, rb_id=0, ignore_warning=False):
        """
        Rollback the candidate config to either the last active or
        a specific rollback number.

        :param int rb_id: The rollback id value [0-49], defaults to ``0``.

        :param ignore_warning: A boolean, string or list of string.
          If the value is True, it will ignore all warnings regardless of the
          warning message. If the value is a string, it will ignore
          warning(s) if the message of each warning matches the string. If
          the value is a list of strings, ignore warning(s) if the message of
          each warning matches at least one of the strings in the list.

        :returns:
            ``True`` always when successful

        :raises ValueError: When invalid rollback id is given
        """

        if rb_id < 0 or rb_id > 49:
            raise ValueError("Invalid rollback #" + str(rb_id))

        self.rpc.load_configuration(
            dict(compare="rollback", rollback=str(rb_id)), ignore_warning=ignore_warning
        )

        return True

    # -------------------------------------------------------------------------
    # rescue configuration
    # -------------------------------------------------------------------------

    def rescue(self, action, format="text"):
        """
        Perform action on the "rescue configuration".

        :param str action: identifies the action as follows:

            * "get" - retrieves/returns the rescue configuration via **format**
            * "save" - saves current configuration as rescue
            * "delete" - removes the rescue configuration
            * "reload" - loads the rescue config as candidate (no-commit)

        :param str format: identifies the return format when **action** is
                           "get":

            * "text" (default) - ascii-text format
            * "xml" - as XML object

        :return:

            * When **action** is 'get', then the contents of the rescue
              configuration is returned in the specified *format*.  If there
              is no rescue configuration saved, then the return value is
              ``None``.

            * ``True`` when **action** is "save".

            * ``True`` when **action** is "delete".

            .. note:: ``True`` regardless if a rescue configuration exists.

            * When **action** is 'reload', return is ``True`` if a rescue
              configuration exists, and ``False`` otherwise.

            .. note:: The rescue configuration is only loaded as the candidate,
                      and not committed.  You must commit to make the rescue
                      configuration active.

        :raises ValueError:
            If **action** is not one of the above
        """

        def _rescue_save():
            """
            Saves the current configuration as the rescue configuration
            """
            self.rpc.request_save_rescue_configuration()
            return True

        def _rescue_delete():
            """
            Deletes the existing resuce configuration.
            """
            # note that this will result in an "OK" regardless if
            # a rescue config exists or not.
            self.rpc.request_delete_rescue_configuration()
            return True

        def _rescue_get():
            """
            Retrieves the rescue configuration, returning it in
            either :format: 'text' or 'xml'.

            Returns either the 'text'/'xml' if the rescue config
            exists, or :None: otherwise
            """
            try:
                got = self.rpc.get_rescue_information(format=format)
                return (
                    got.findtext("configuration-information/configuration-output")
                    if "text" == format
                    else got
                )
            except:
                return None

        def _rescue_reload():
            """
            Loads the rescue configuration as the active candidate.
            This action does *not* commit the configuration; use the
            :commit(): method for that purpose.

            Returns the XML response if the rescue configuration
            exists, or :False: otherwise
            """
            try:
                return self.rpc.load_configuration({"rescue": "rescue"})
            except:
                return False

        def _unsupported_action():
            raise ValueError("unsupported action: {}".format(action))

        result = {
            "get": _rescue_get,
            "save": _rescue_save,
            "delete": _rescue_delete,
            "reload": _rescue_reload,
        }.get(action, _unsupported_action)()

        return result

    def __init__(self, dev, mode=None, **kwargs):
        """
        :param str mode: Can be used *only* when creating Config object using
                         context manager

            * "private" - Work in private database
            * "dynamic" - Work in dynamic database
            * "batch" - Work in batch database
            * "exclusive" - Work with Locking the candidate configuration
            * "ephemeral" - Work in default/specified ephemeral instance

        :param str ephemeral_instance: ephemeral instance name

        .. code-block:: python

           # mode can be private/dynamic/exclusive/batch/ephemeral
           with Config(dev, mode='exclusive') as cu:
               cu.load('set system services netconf traceoptions file xyz',
                       format='set')
               print cu.diff()
               cu.commit()

        .. warning::
            Ephemeral databases are an advanced Junos feature which
            if used incorrectly can have serious negative impact on the operation
            of the Junos device. We recommend you consult JTAC and/or you Juniper
            account team before deploying the ephemeral database feature in your
            network.
        """
        self.mode = mode
        if not kwargs.get("ephemeral_instance") and kwargs:
            raise ValueError("Unsupported argument provided to Config class")
        self.kwargs = kwargs

        Util.__init__(self, dev=dev)

    def __enter__(self):
        # defining separate functions for each mode so that can be
        # changed/edited as per the need of corresponding rpc call.
        def _open_configuration_private():
            try:
                self.rpc.open_configuration(private=True)
            except (RpcTimeoutError, ConnectClosedError) as err:
                raise err
            except RpcError as err:
                if err.rpc_error["severity"] == "warning":
                    if (
                        err.message != "uncommitted changes will be discarded "
                        "on exit"
                    ):
                        warnings.warn(err.message, RuntimeWarning)
                    return True
                else:
                    raise err

        def _open_configuration_dynamic():
            try:
                self.rpc.open_configuration(dynamic=True)
            except RpcError as err:
                raise err
            return True

        def _open_configuration_batch():
            try:
                self.rpc.open_configuration(batch=True)
            except (RpcTimeoutError, ConnectClosedError) as err:
                raise err
            except RpcError as err:
                if err.rpc_error["severity"] == "warning":
                    if (
                        err.message != "uncommitted changes will be discarded "
                        "on exit"
                    ):
                        warnings.warn(err.message, RuntimeWarning)
                    return True
                else:
                    raise err

        def _open_configuration_exclusive():
            return self.lock()

        def _open_configuration_ephemeral(**kwargs):
            self.rpc.open_configuration(**kwargs)
            return True

        def _unsupported_option():
            if self.mode is not None:
                raise ValueError("unsupported action: {}".format(self.mode))

        if self.kwargs.get("ephemeral_instance"):
            ephemeral_kwargs = {"ephemeral_instance": self.kwargs["ephemeral_instance"]}
        else:
            ephemeral_kwargs = {"ephemeral": True}

        {
            "private": _open_configuration_private,
            "dynamic": _open_configuration_dynamic,
            "batch": _open_configuration_batch,
            "exclusive": _open_configuration_exclusive,
            "ephemeral": lambda: _open_configuration_ephemeral(**ephemeral_kwargs),
        }.get(self.mode, _unsupported_option)()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.mode == "exclusive":
            self.unlock()
        elif self.mode is not None:
            self.rpc.close_configuration()



from jnpr.junos.exception import RpcError
import re


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "current_re": "A list of internal routing instance hostnames "
        "for the current RE. These hostnames identify "
        "things like the RE's slot ('re0' or 're1'), the "
        "RE's mastership state ('master' or 'backup'), "
        "and node in a VC ('member0' or 'member1')",
    }


def get_facts(device):
    """
    The RPC-equivalent of show interfaces terse on private routing instance.
    """
    current_re = []

    try:
        if device.facts["_is_linux"]:
            rsp = device.rpc.get_system_nodes_information(normalize=True)

            # Get the RE (Active) list from the response
            res_list = rsp.xpath(
                "system-nodes-info-entry/system-node-info-node-attributes-list/"
                "system-node-info-node-attribute[text()='RE (Active)']/../.."
            )
            for re_ele in res_list:
                # Use the node-name element ot get RE name
                re_name = re_ele.xpath("system-node-info-node-name")
                current_re.append(re_name[0].text)
        else:
            rsp = device.rpc.get_interface_information(
                normalize=True,
                routing_instance="__juniper_private1__",
                terse=True,
            )

            # Get the local IPv4 addresses from the response.
            for ifa in rsp.iterfind(
                ".//address-family[address-family-name='inet']/"
                "interface-address/ifa-local"
            ):
                ifa_text = ifa.text
                if ifa_text is not None:
                    # Separate the IP from the mask
                    (ip, _, _) = ifa.text.partition("/")
                    if ip is not None:
                        # Use the _iri_hostname fact to map the IP address to
                        # an internal routing instance hostname.
                        if ip in device.facts["_iri_hostname"]:
                            for host in device.facts["_iri_hostname"][ip]:
                                if host not in current_re:
                                    current_re.append(host)
                        # An SRX platform in an HA cluster uses a different
                        # algorithm for assigning IRI IP addresses
                        elif device.facts["srx_cluster_id"] is not None:
                            try:
                                # Split the IRI IP into a list of 4 octets
                                octets = ip.split(".", 3)
                                # The 2nd octet will be cluster-id << 4
                                cluster_id_octet = str(
                                    (int(device.facts["srx_cluster_id"]) & 0x000F) << 4
                                )
                                if cluster_id_octet == "0":
                                    cluster_id_octet = str(
                                        device.facts["srx_cluster_id"]
                                    )
                                # node0 will have an IP of
                                #     129.<cluster_id_octet>.0.1
                                # node1 will have an IP of
                                #     130.<cluster_id_octet>.0.1
                                # primary will have an IP of
                                #     143.<cluster_id_octet>.0.1
                                if (
                                    octets[1] == cluster_id_octet
                                    and octets[2] == "0"
                                    and octets[3] == "1"
                                ):
                                    host = None
                                    if octets[0] == "129":
                                        host = "node0"
                                    elif octets[0] == "130":
                                        host = "node1"
                                    elif octets[0] == "143":
                                        host = "primary"
                                    if host is not None and host not in current_re:
                                        current_re.append(host)
                            # Problem splitting IP into octets and indexing them.
                            # Keep looping to check the other IRI IPs.
                            except IndexError:
                                pass
    except RpcError:
        # Check to see if it's JDM of Junos Node slicing
        try:
            current_re_sw = device.rpc.get_software_information()
            if current_re_sw is not None:
                server_slot = current_re_sw.findtext(
                    './package-information[name="Server ' 'slot"]/comment'
                )
                slot_num = re.findall(r"Server slot : (\d+)", server_slot)[0]
                current_re = ["server" + slot_num]
        except Exception:
            pass

    # An empty list indicates a problem finding any current_re info.
    # Return None.
    if len(current_re) == 0:
        current_re = None

    return {
        "current_re": current_re,
    }



from collections import OrderedDict
from functools import reduce
import re
import copy
import logging

from jinja2 import Template, meta
from transitions import Machine
import pyparsing as pp

logger = logging.getLogger("jnpr.junos.factory.state_machine")


class Identifiers:
    """
    This class static variables can be used when defining regex view. For ex:
    _PFENotificationStatsTable:
      title: PFE Notification statistics
      key: name
      view: _PFENotificationStatsView

    _PFENotificationStatsView:
      regex:
        value: numbers
        name: words
    """

    printables = pp.OneOrMore(pp.Word(pp.printables))
    numbers = (
        pp.Word(pp.nums) + pp.Optional(pp.Literal(".") + pp.Word(pp.nums))
    ).setParseAction(lambda i: "".join(i))
    hex_numbers = pp.OneOrMore(pp.Word(pp.nums, min=1)) & pp.OneOrMore(
        pp.Word("abcdefABCDEF", min=1)
    )
    word = pp.Word(pp.alphanums) | pp.Word(pp.alphas)
    words = (pp.OneOrMore(word)).setParseAction(lambda i: " ".join(i))
    percentage = pp.Word(pp.nums) + pp.Literal("%")
    header_bar = (
        pp.OneOrMore(pp.Word("-")) | pp.OneOrMore(pp.Word("="))
    ) + pp.StringEnd()


def data_type(item):
    """

    Args:
        item: string element parsed from the blob.

    Returns: item converted to data type it should represent

    """
    try:
        obj = Identifiers.numbers.parseString(item, parseAll=True)
        if "." in obj[0]:
            return float
        else:
            return int
    except pp.ParseException as ex:
        pass
    try:
        Identifiers.hex_numbers.parseString(item, parseAll=True)
        return str  # special case
    except pp.ParseException as ex:
        pass
    return str


def convert_to_data_type(items):
    """

    Args:
        items: list of string objects

    Returns: list of converted data type objects

    """
    item_types = map(data_type, items)
    return list(map(lambda x, y: int(x) if y is int else x.strip(), items, item_types))


class StateMachine(Machine):
    def __init__(self, table_view):
        self._data = {}
        self._table = table_view
        self._view = self._table.VIEW
        self._raw = ""
        self._lines = []
        self.states = [
            "row_column",
            "title_data",
            "regex_data",
            "delimiter_data",
            "exists_bool_data",
        ]
        self.transitions = [
            {
                "trigger": "column_provided",
                "source": "*",
                "dest": "row_column",
                "conditions": "match_columns",
                "before": "check_header_bar",
                "after": "parse_raw_columns",
            },
            {
                "trigger": "check_next_row",
                "source": "row_column",
                "dest": "row_column",
                "conditions": "prev_next_row_same_type",
                "after": "parse_raw_columns",
            },
            {
                "trigger": "title_provided",
                "source": "start",
                "dest": "title_data",
                "conditions": ["match_title", "title_not_followed_by_columns"],
                "after": "parse_title_data",
            },
            {
                "trigger": "regex_provided",
                "source": "title_data",
                "dest": "regex_data",
                "conditions": ["match_title"],
                "before": "check_header_bar",
                "after": "parse_using_regex",
            },
            {
                "trigger": "delimiter_without_title",
                "source": "start",
                "dest": "delimiter_data",
                "after": "parse_using_delimiter",
            },
            {
                "trigger": "delimiter_with_title",
                "source": ["start", "delimiter_data"],
                "dest": "delimiter_data",
                "conditions": ["match_title"],
                "before": "check_header_bar",
                "after": "parse_using_delimiter",
            },
            {
                "trigger": "regex_with_item",
                "source": ["start", "regex_data"],
                "dest": "regex_data",
                "after": "parse_using_item_and_regex",
            },
            {
                "trigger": "regex_parser",
                "source": "start",
                "dest": "regex_data",
                "after": "parse_using_regex",
            },
            {
                "trigger": "regex_parser",
                "source": "regex_data",
                "dest": "regex_data",
                "after": "parse_using_regex",
            },
            {
                "trigger": "regex_parser",
                "source": "row_column",
                "dest": "regex_data",
                "after": "parse_using_regex",
            },
            {
                "trigger": "exists_check",
                "source": ["start", "regex_data", "row_column"],
                "dest": "exists_bool_data",
                "after": "parse_exists",
            },
            {
                "trigger": "exists_check",
                "source": "title_data",
                "dest": "exists_bool_data",
                "after": "parse_exists",
            },
        ]
        Machine.__init__(
            self,
            states=self.states,
            transitions=self.transitions,
            initial="start",
            send_event=True,
        )

    def parse(self, lines):
        """
        Starting point to parse string blob section. In case of nested table
        this API will get called in recursive call.

        Args:
            lines: list of lines which was received from rpc reply <output>.

        Returns: dictionary (self._data) with parsed data.

        """
        self._lines = copy.deepcopy(lines)
        self._raw = "\n".join(lines)
        if self._view is None:
            if self._table.DELIMITER is not None:
                if self._table.TITLE is not None:
                    self.delimiter_with_title()
                else:
                    self.delimiter_without_title()
            elif self._table.TITLE is not None:
                self.title_provided()
            elif self._table.ITEM is not None and self._table.ITEM != "*":
                self._parse_item_iter(lines)
        else:
            if self._view.TITLE is not None or self._table.TITLE:
                self.title_provided()
            if len(self._view.REGEX) > 0:
                if self._table.ITEM is not None:
                    self.regex_with_item()
                else:
                    self.regex_parser()
            if len(self._view.COLUMNS) > 0:
                self.column_provided()
            if len(self._view.EXISTS) > 0:
                self.exists_check()
            if len(self._view.FIELDS) > 0:
                if self._table.ITEM is not None and self._table.ITEM != "*":
                    return self._parse_item_iter(lines)
                for key, value in self._view.FIELDS.items():
                    tbl = value["table"]
                    tbl._view = tbl.VIEW
                    if tbl._view is None:
                        self._data[key] = StateMachine(tbl).parse(lines)
                        continue
                    if len(tbl._view.COLUMNS) > 0:
                        self._data[key] = StateMachine(tbl).parse(lines)
                    if tbl._view.TITLE is not None or tbl.TITLE is not None:
                        self._data[key] = StateMachine(tbl).parse(lines)
        if self._table.EVAL:
            self._eval_in_full_data()
        return self._data

    def _eval_in_full_data(self):
        """
        To eval expression from full set of data (self._data). Sets key in
        self._data itself.

        Returns: None
        """
        for name, expression in self._table.EVAL.items():
            t = Template(expression)
            expression = t.render(data=self._data)
            try:
                val = eval(expression)
            except Exception as ex:
                logger.error("eval expression for '%s' failed due to %s" % (name, ex))
                self._data[name] = None
            else:
                self._data[name] = val

    def _parse_item_iter(self, lines):
        """
        There are cases when similar data set repeats. With item its easier to
        split string blob into different sections.
        Args:
            lines: list of lines

        Returns: dictionary (self._data) with parsed data.

        """
        self._raw = "\n".join(lines)
        pat = pp.Regex(self._table.ITEM)
        x = []
        for result, start, end in pat.scanString(self._raw):
            x.append((start, end))
        for i in range(len(x)):
            try:
                raw = self._raw[x[i][0] : x[i + 1][0]]
            except IndexError:
                raw = self._raw[x[i][0] :]
            lines = raw.splitlines()
            obj = re.search(self._table.ITEM, lines[0])
            groups = obj.groups()
            groups = [data_type(val)(val) for val in groups]
            if len(groups) >= 1:
                if len(groups) != len(
                    self._table.KEY
                    if isinstance(self._table.KEY, list)
                    else [self._table.KEY]
                ):
                    raise KeyError(
                        "Table with grouped item must contain " "corresponding key(s)"
                    )
                master_key = groups[0] if len(groups) == 1 else tuple(groups)
                self._data[master_key] = {}
                if self._view is not None:
                    for key, value in self._view.FIELDS.items():
                        tbl = value["table"]
                        tbl._view = tbl.VIEW
                        if tbl._view is not None and len(tbl._view.COLUMNS) > 0:
                            self._data[master_key][key] = StateMachine(tbl).parse(lines)
                        if tbl.TITLE is not None or tbl._view.TITLE is not None:
                            self._data[master_key][key] = StateMachine(tbl).parse(lines)
                else:
                    self._table.TITLE = lines[0]
                    delimiter = self._table.DELIMITER or r"\s\s+"
                    temp_dict = {}
                    pre_space_delimit = self._get_pre_space_delimiter(lines[1])
                    for line in lines[1:]:
                        if re.match(pre_space_delimit + r"\s+", line):
                            break
                        if line.startswith(pre_space_delimit):
                            try:
                                items = re.split(delimiter, line.strip())
                                item_types = list(map(data_type, items))
                                key, value = convert_to_data_type(items)
                                temp_dict[key] = value
                            except ValueError:
                                regex = (
                                    r"(\d+)\s(.*)"
                                    if item_types[0] == int
                                    else "(" r".*)\s(\d+)"
                                )
                                obj = re.search(regex, line)
                                if obj:
                                    items = obj.groups()
                                    key, value = convert_to_data_type(items)
                                    temp_dict[key] = value
                        # check if next line is blank or new title (delimiter
                        # test to fail)
                        elif (
                            line.strip() == ""
                            or len(re.split(delimiter, line.strip())) <= 1
                        ):
                            break
                    self._insert_eval_data(temp_dict)
                    self._data[master_key] = temp_dict
        return self._data

    def match_columns(self, event):
        if self._view is None:
            return False
        columns = self._view.COLUMNS.values()
        if len(columns) == 0:
            return False
        if True in [isinstance(i, list) for i in columns]:
            # we have multi line column header
            columns = [[i] if isinstance(i, str) else i for i in columns]
            max_title_len = reduce(
                lambda x, y: x if x > y else y, map(lambda x: len(x), columns)
            )
            for index, item in enumerate(columns):
                columns[index] = item + [None] * (max_title_len - len(item))
            col_parser = reduce(lambda x, y: x & y, [pp.Literal(i[0]) for i in columns])
            for line in self._lines:
                if self._parse_literal(line, col_parser):
                    for index in range(1, max_title_len):
                        col_parser = reduce(
                            lambda x, y: x & y,
                            [
                                pp.Literal(i[index])
                                for i in columns
                                if i[index] is not None
                            ],
                        )
                        if self._parse_literal(
                            self._lines[self._lines.index(line) + 1], col_parser
                        ):
                            # removing next lines in header title, dont see any
                            # use of these lines
                            self._lines.pop(self._lines.index(line) + 1)
                        else:
                            return False
                    current_index = self._lines.index(line)
                    self._lines = self._lines[current_index:]
                    return True
                else:
                    continue
        else:
            col_parser = reduce(lambda x, y: x & y, [pp.Literal(i) for i in columns])
            for line in self._lines:
                if self._parse_literal(line, col_parser):
                    d = set(map(lambda x, y: x in y, columns, [line] * len(columns)))
                    if d.pop():
                        current_index = self._lines.index(line)
                        self._lines = self._lines[current_index:]
                        return True
            return False

    def match_title(self, event):
        """
        To check whether given title exists in string blob.

        Args:
            event: In case any data is supposed to be passed.

        Returns: Boolean value depending on matching of title

        """
        title = self._table.TITLE or self._view.TITLE
        for line in self._lines:
            if title in line:
                current_index = self._lines.index(line)
                self._lines = self._lines[current_index:]
                return True
        return False

    def title_not_followed_by_columns(self, event):
        return not self.match_columns(event)

    def _parse_literal(self, line, col_parser):
        try:
            if col_parser.searchString(line):
                return True
        except pp.ParseException as ex:
            return False

    def check_header_bar(self, event):
        """
        Remove redundant title bar line from lines.
        Args:
            event:

        Returns: Boolean depending on if the next line is a title bar or not.

        For example, here if we provide columns, we need to get rid of next line
        ---------------------------------------
        Module  Name              Active Errors
        ---------------------------------------
        1       PQ3 Chip          0
        2       Host Loopback     0

        """
        line = self._lines[1]
        try:
            Identifiers.header_bar.parseString(line, parseAll=True)
            self._lines.pop(1)
        except pp.ParseException as ex:
            return False
        return True

    def parse_raw_columns(self, event):
        col_offsets = {}
        col_order = event.kwargs.get("col_order", OrderedDict())
        line = self._lines[0]
        column = self._view.COLUMNS
        if True in [isinstance(i, list) for i in column.values()]:
            # just take first line of column title
            for k, v in column.items():
                if isinstance(v, list):
                    column[k] = v[0]
        if len(col_order) == 0:
            for key, val in column.items():
                for result, start, end in pp.Literal(val).scanString(line):
                    col_offsets[(start, end)] = result[0]
            user_defined_columns = copy.deepcopy(self._view.COLUMNS)
            for key in sorted(col_offsets.keys()):
                for x, y in self._view.COLUMNS.items():
                    if col_offsets[key] == user_defined_columns.get(x):
                        col_order[key] = x
                        user_defined_columns.pop(x)
                        break
        key = self._get_key(event.kwargs.get("key", self._table.KEY))
        items = re.split(r"\s\s+", self._lines[1].strip())

        post_integer_data_types = event.kwargs.get("check", list(map(data_type, items)))
        index = event.kwargs.get("index", 1)
        # col_len = len(col_order)
        columns_list = list(col_order.values())
        for index, line in enumerate(self._lines[index:], start=index):
            items = re.split(r"\s\s+", line.strip())
            if len(items) >= len(columns_list):
                if len(items) > len(columns_list):
                    if (
                        list(col_offsets.keys())[0][0] > 10
                        and self._table.KEY == "name"
                    ):
                        columns_list.insert(0, self._table.KEY)
                    else:
                        items = items[: len(columns_list)]
                post_integer_data_types, pre_integer_data_types = (
                    list(map(data_type, items)),
                    post_integer_data_types,
                )
                if post_integer_data_types == pre_integer_data_types:
                    items = list(
                        map(lambda data, typ: typ(data), items, post_integer_data_types)
                    )
                    tmp_dict = dict(zip(columns_list, items))
                    self._insert_data(key, tmp_dict, columns_list)
                else:
                    break
            elif line.strip() == "":
                # first check if loop already reached at the end of command o/p
                if not index + 1 >= len(self._lines):
                    self.check_next_row(
                        check=post_integer_data_types,
                        data=self._data,
                        index=index,
                        col_order=col_order,
                        key=key,
                    )
        return self._data

    def _insert_data(self, key, tmp_dict, columns_list):
        """
        Insert data per row into main dictionary self._data

        Args:
            key: To be used as key to dictionary
            tmp_dict: Temporary dictionary to be populated
            columns_list: Columns defined by the users

        Returns: None, function just populates self._data

        """
        self._insert_eval_data(tmp_dict)
        if isinstance(key, (tuple, list)):
            if self._view.FILTERS is not None:
                selected_dict = {}
                for select in self._view.FILTERS:
                    if select in columns_list:
                        selected_dict[select] = tmp_dict[select]
                if self._table.KEY_ITEMS is None:
                    self._data[tuple(tmp_dict[i] for i in key)] = selected_dict
                elif tmp_dict[key] in self._table.KEY_ITEMS:
                    self._data[tuple(tmp_dict[i] for i in key)] = selected_dict
            else:
                self._data[tuple(tmp_dict[i] for i in key)] = tmp_dict
        else:
            if self._view.FILTERS is not None:
                selected_dict = {}
                for select in self._view.FILTERS:
                    if select in columns_list:
                        selected_dict[select] = tmp_dict[select]
                if self._table.KEY_ITEMS is None:
                    self._data[tmp_dict[key]] = selected_dict
                elif tmp_dict[key] in self._table.KEY_ITEMS:
                    self._data[tmp_dict[key]] = selected_dict
            else:
                if self._table.KEY_ITEMS is None:
                    if key not in tmp_dict:
                        self._data.update(tmp_dict)
                    else:
                        self._data[tmp_dict[key]] = tmp_dict
                elif tmp_dict[key] in self._table.KEY_ITEMS:
                    self._data[tmp_dict[key]] = tmp_dict

    def _get_key(self, key):
        """
        Fetch the key from columns which will be used to define dictionary
        Args:
            key: user defined key input

        Returns: key to be used in dictionary.

        """
        if isinstance(key, list):
            if set(
                [
                    i in self._view.COLUMNS or i in self._view.COLUMNS.values()
                    for i in key
                ]
            ):
                key_temp = []
                for i in key:
                    if i not in self._view.COLUMNS and i in self._view.COLUMNS.values():
                        for user_provided, from_table in self._view.COLUMNS.items():
                            # as dict will be created with user_provided key
                            if i == from_table or i == user_provided:
                                key_temp.append(user_provided)
                                break
                    else:
                        key_temp.append(i)
                key = tuple(key_temp)
        elif key not in self._view.COLUMNS and key in self._view.COLUMNS.values():
            for user_provided, from_table in self._view.COLUMNS.items():
                if key == from_table:
                    key = user_provided
        key = key[0] if len(key) == 1 and isinstance(key, tuple) else key
        return key

    def prev_next_row_same_type(self, event):
        """
        Checks if the consecutive two lines of similar types.

        Args:
            event: takes parameter from the API call

        Returns: Boolean, depending if two lines are of similar types.

        """
        index = event.kwargs.get("index")
        post_integer_data_types = event.kwargs.get("check")
        line = self._lines[index]
        items = re.split(r"\s\s+", line.strip())
        post_integer_data_types, pre_integer_data_types = (
            list(map(data_type, items)),
            post_integer_data_types,
        )
        return post_integer_data_types == pre_integer_data_types

    def _get_pre_space_delimiter(self, line):
        """
        Sometime all the data lines under a table is withing some space limit
        which gives a easy way to stop looking for lines if the delimiter is not
        followed.
        Args:
            line: should be first line after title

        Returns: decide how much space to be considered as delimiter

        For example, in below output, all data for FPCx has 2 space delimiter

        Statistics for port 1 connected to device FPC1:
          TX Packets 64 Octets        96106518
          TX Packets 65-127 Octets    56224217
          TX Packets 128-255 Octets   4550198
          TX Packets 256-511 Octets   1841
          TX Packets 512-1023 Octets  1354
        Statistics for port 1 connected to device FPC2:
          TX Packets 64 Octets        345354435
          TX Packets 65-127 Octets    34531
          TX Packets 128-255 Octets   4351451
          TX Packets 256-511 Octets   1345
          TX Packets 512-1023 Octets  526513
        """
        pre_space_delimit = ""
        obj = re.search(r"(\s+).*", line)
        if obj:
            pre_space_delimit = obj.group(1)
        return pre_space_delimit

    def parse_title_data(self, event):
        """
        title description in Table helps to get the starting point for starting
        search

        Args:
            event: In case trigger want to pass some data

        Returns: None, Just add corresponding key, value to existing dict

        Let say we have CLI "show xmchip 0 pt stats" output as:

        WAN PT statistics (Index 0)
        ---------------------------

        PCT entries used by all WI-1 streams         : 0
        PCT entries used by all WI-0 streams         : 0
        PCT entries used by all LI streams           : 0
        CPT entries used by all multicast packets    : 0
        CPT entries used by all WI-1 streams         : 0
        CPT entries used by all WI-0 streams         : 0
        CPT entries used by all LI streams           : 0

        Fabric PT statistics (Index 1)
        ------------------------------

        PCT entries used by all FI streams           : 0
        PCT entries used by all WI (Unused) streams  : 0
        PCT entries used by all LI streams           : 0
        CPT entries used by all multicast packets    : 0
        CPT entries used by all FI streams           : 0
        CPT entries used by all WI (Unused) streams  : 0
        CPT entries used by all LI streams           : 0

        With below table/view to parse the data
        ---
        XMChipStatsTable:
          command: show xmchip 0 pt stats
          target: fpc1
          view: XMChipStatsView

        XMChipStatsView:
          fields:
            wan_pt_stats: _WANPTStatTable
            fabric_pt_stats: _FabricPTStatTable

        _WANPTStatTable:
          title: WAN PT statistics (Index 0)
          delimiter: ":"

        _FabricPTStatTable:
          title: Fabric PT statistics (Index 1)
          delimiter: ":"

        which returns:
        {'fabric_pt_stats': {'CPT entries used by all FI streams': 0,
                     'CPT entries used by all LI streams': 0,
                     'CPT entries used by all WI (Unused) streams': 0,
                     'CPT entries used by all multicast packets': 0,
                     'PCT entries used by all FI streams': 0,
                     'PCT entries used by all LI streams': 0,
                     'PCT entries used by all WI (Unused) streams': 0},
        'wan_pt_stats': {'CPT entries used by all LI streams': 0,
                     'CPT entries used by all WI-0 streams': 0,
                     'CPT entries used by all WI-1 streams': 0,
                     'CPT entries used by all multicast packets': 0,
                     'PCT entries used by all LI streams': 0,
                     'PCT entries used by all WI-0 streams': 0,
                     'PCT entries used by all WI-1 streams': 0}}
        """
        if self._view is not None and self._view.REGEX != {}:
            return self.regex_provided()
        # view have only fields, so contains only nested table
        if (
            self._view is not None
            and self._view.FIELDS
            and not self._view.COLUMNS
            and not self._view.REGEX
        ):
            return
        delimiter = self._table.DELIMITER or r"\s\s+"
        pre_space_delimit = self._get_pre_space_delimiter(self._lines[1])
        for line in self._lines[1:]:
            if re.match(pre_space_delimit + r"\s+", line):
                break
            if line.startswith(pre_space_delimit):
                try:
                    items = re.split(delimiter, line.strip())
                    item_types = list(map(data_type, items))
                    key, value = convert_to_data_type(items)
                    if self._view is None:
                        self._data[key] = value
                    elif self._table.KEY_ITEMS is None:
                        self._data[self._view.FIELDS.get(key, key)] = value
                    elif key in self._table.KEY_ITEMS:
                        self._data[self._view.FIELDS.get(key, key)] = value
                except ValueError:
                    regex = (
                        r"(\d+)\s(.*)" if item_types[0] == int else "(" r".*)\s(\d+)"
                    )
                    obj = re.search(regex, line)
                    if obj:
                        items = obj.groups()
                        key, value = convert_to_data_type(items)
                        self._data[key] = value
            # check if next line is blank or new title (delimiter test to fail)
            elif line.strip() == "" or len(re.split(delimiter, line.strip())) <= 1:
                break
        return self._data

    def parse_using_regex(self, event):
        r"""
        All the regex should add up to match a line

        Args:
            event: In case trigger want to pass some data

        Returns: None, Just add corresponding key, value to existing dict

        Let say we have CLI "show system processes extensive" output as:

        PID USERNAME  HR PRI NICE   SIZE    RES STATE   C   TIME    WCPU COMMAND
           11 root       2 155 ki31     0K    32K RUN     0 3214.6 191.31% idle
        19542 root       2  22    0   808M 49872K select  1  90.9H   3.86% chas
        19679 root       1  33    0   723M 1208K select  0 839:37   0.49% python
        19451 root       1  20    0   720M 1052K select  1  60:37   0.20% eventd
           12 root       26 -60    -     0K   416K WAIT    1 635:29   0.00% intr

        With a regex table as below
        ---
        SystemProcExtTable:
          command: show system processes extensive
          key: cmd
          view: SystemProcExtView

        SystemProcExtView:
          regex:
            pid: '\d+'
            wcpu: '.*(\d+\.\d+)%'
            cmd: '\w+'

        which returns something like (data for representation only):
        {'alarm': {'cmd': 'alarm', 'pid': 20483, 'wcpu': '0.00'},
         'alarmd': {'cmd': 'alarmd', 'pid': 20473, 'wcpu': '0.00'},
         'appidd': {'cmd': 'appidd', 'pid': 19686, 'wcpu': '0.00'},
         'apsd': {'cmd': 'apsd', 'pid': 20435, 'wcpu': '0.00'},
        """
        _regex = copy.deepcopy(self._view.REGEX)
        for key, val in _regex.items():
            if val in Identifiers.__dict__:
                _regex[key] = Identifiers.__dict__[val]
            else:
                _regex[key] = pp.Regex(val, flags=re.IGNORECASE)

        _regex = reduce(lambda x, y: x + y, _regex.values())
        for index, line in enumerate(self._lines):
            tmp_dict = {}
            # checking index as there can be blank line at position 0 and 2
            if line.strip() == "":
                if index > 2:
                    if self.is_row_column() or self.is_regex_data():
                        try:
                            match = []
                            # check if line after new line matches condition
                            # There can be data set where there are new line in between.
                            for result, start, end in _regex.scanString(
                                self._lines[index + 1]
                            ):
                                match.extend(result)
                            if match:
                                continue
                            else:
                                break
                        except IndexError:
                            break
                else:
                    continue
            for result, start, end in _regex.scanString(line):
                # write a different function for this
                for key, val in self._view.REGEX.items():
                    if val not in Identifiers.__dict__:
                        obj = re.search(
                            val, result[list(self._view.REGEX.keys()).index(key)], re.I
                        )
                        if obj and len(obj.groups()) >= 1:
                            result[
                                list(self._view.REGEX.keys()).index(key)
                            ] = obj.groups()[0]
                items = convert_to_data_type(result)
                tmp_dict = dict(zip(self._view.REGEX.keys(), items))
                if len(tmp_dict) > 0:
                    self._insert_data(
                        self._table.KEY, tmp_dict, list(self._view.REGEX.keys())
                    )

    def parse_using_item_and_regex(self, event):
        r"""
        when multiple map is provided for regex, they gets added to search each
        line. But when item is '*' each regex item is used to search given value
        regular expression in whole string blob.

        Args:
            event: In case trigger want to pass some data

        Returns: None, Just add corresponding key, value to existing dict

        Let say we have CLI "show xmchip 0 pt stats" output as:

        PCT entries used by all WI-1 streams         : 21
        PCT entries used by all WI-0 streams         : 34
        PCT entries used by all LI streams           : 0
        CPT entries used by all multicast packets    : 0
        CPT entries used by all WI-1 streams         : 0
        CPT entries used by all WI-0 streams         : 0
        CPT entries used by all LI streams           : 0

        ---
        XMChipStatsTable:
          command: show xmchip {{ instance }} pt stats
          args:
            instance: 0
          target: fpc2
          item: '*'
          view: XMChipStatsView

        XMChipStatsView:
          regex:
            pct_wi_1: 'PCT entries used by all WI-1 streams\s+:\s?(\d+)'
            pct_wi_0: 'PCT entries used by all WI-0 streams\s+:\s?(\d+)'

        which returns:
        {'pct_wi_1': 0, 'pct_wi_0': 0}
        """
        if self._table.ITEM == "*":
            self._raw = "\n".join(self._lines)
            for key, regex in self._view.REGEX.items():
                obj = re.search(regex, self._raw)
                if obj:
                    val = obj.groups()[0] if len(obj.groups()) >= 1 else obj.group()
                    self._data[key] = data_type(val)(val)
            self._insert_eval_data(self._data)
        else:
            key = self._get_key(event.kwargs.get("key", self._table.KEY))
            for line in re.finditer("%s .*" % self._table.ITEM, "\n".join(self._lines)):
                tmp_dict = {}
                for k, exp in self._view.REGEX.items():
                    obj = re.search(exp, line.group())
                    if obj:
                        val = obj.groups()[0] if len(obj.groups()) >= 1 else obj.group()
                        tmp_dict[k] = data_type(val)(val)
                self._insert_eval_data(tmp_dict)
                if key in tmp_dict:
                    self._data[tmp_dict[key]] = tmp_dict
        return self._data

    def parse_using_delimiter(self, event):
        """

        Args:
            event: In case trigger want to pass some data

        Returns: None, Just add corresponding key, value to existing dict

        For a given cli "show link stats" output:
        PPP LCP/NCP: 0
        HDLC keepalives: 0
        RSVP: 0
        ISIS: 0
        OSPF Hello: 541025
        OAM:  0
        BFD:  15
        UBFD:  0
        LMI:  0
        LACP: 0
        ETHOAM: 0
        SYNCE:  0
        PTP:  0
        L2TP:  0
        LNS-PPP:  0
        ARP:  4307
        ELMI:  0
        VXLAN MRESOLVE: 0
        Unknown protocol: 42

        Lets say we can have table as
        ---
        FPCLinkStatTable:
            command: show link stats
            target: fpc1
            delimiter: ":"

        which given return value as
        {'ARP': 4307, 'ELMI': 0, 'SYNCE': 0, 'ISIS': 0, 'BFD': 15,
        'PPP LCP/NCP': 0, 'OAM': 0, 'ETHOAM': 0, 'LACP': 0, 'LMI': 0,
        'Unknown protocol': 42, 'UBFD': 0, 'L2TP': 0, 'HDLC keepalives': 0,
        'LNS-PPP': 0, 'OSPF Hello': 541025, 'RSVP': 0, 'VXLAN MRESOLVE': 0,
        'PTP': 0}

        """
        delimiter = self._table.DELIMITER or r"\s\s+"
        pre_space_delimit = ""
        if self._table.TITLE is None:
            for line in self._lines[1:]:
                if line.strip() == "SENT: Ukern command: %s" % self._table.GET_CMD:
                    self._lines = self._lines[self._lines.index(line) + 1 :]
                    break
        else:
            obj = re.search(r"^(\s+).*", self._lines[1])
            if obj:
                pre_space_delimit = obj.group(1)
        for index, line in enumerate(self._lines[1:]):
            if line.strip() == "":
                if index > 2:
                    break
                else:
                    continue
            if line.startswith(pre_space_delimit):
                try:
                    items = re.split(delimiter, line.strip())
                    key, value = convert_to_data_type(items)
                    if self._table.KEY_ITEMS is None:
                        self._data[key] = value
                    elif key in self._table.KEY_ITEMS:
                        self._data[key] = value
                except ValueError as ex:
                    # create a class named ParseError
                    raise Exception("Not able to parse line: %s" % line)
            else:
                break
        return self._data

    def parse_exists(self, event):
        """

        Args:
            event: In case trigger want to pass some data

        Returns: None, Just add corresponding key, value to existing dict. Value
         here is a boolean

        ---
        HostlbStatusSummaryTable:
          command: show host_loopback status-summary
          target: Null
          view: HostlbStatusSummaryView

        HostlbStatusSummaryView:
          exists:
            no_detected_wedges: No detected wedges
            no_toolkit_errors: No toolkit errors

        should give something like
        {'no_detected_wedges': True, 'no_toolkit_errors': True}
        """
        for key, search in self._view.EXISTS.items():
            self._data[key] = re.search(search, self._raw, re.I | re.M) is not None

    def _insert_eval_data(self, tmp_dict):
        r"""

        Args:
            tmp_dict: dictionary of key value from a iteration of view

        Returns: if there is any eval expression in view, does the eval using
        tmp_dictionary. For example:

        ---
        XMChipStatsTable:
          command: show xmchip {{ instance }} pt stats
          args:
            instance: 0
          target: fpc2
          item: '*'
          view: XMChipStatsView

        XMChipStatsView:
          regex:
            pct_wi_1: 'PCT entries used by all WI-1 streams\s+:\s?(\d+)'
            pct_wi_0: 'PCT entries used by all WI-0 streams\s+:\s?(\d+)'
          eval:
            total_pct: '{{ pct_wi_1 }} + {{ pct_wi_0 }}'

        which returns
        {'pct_wi_1': 0, 'pct_wi_0': 0, 'total_pct': 0}
        """
        if self._view and len(self._view.EVAL) > 0:
            for name, expression in self._view.EVAL.items():
                variables = meta.find_undeclared_variables(expression)
                t = Template(expression)
                expression = t.render({k: tmp_dict.get(k) for k in variables})
                val = eval(expression)
                tmp_dict[name] = val



from time import sleep
import logging

from jnpr.junos import exception as EzErrors
from jnpr.junos.transport.tty_netconf import tty_netconf

logger = logging.getLogger("jnpr.junos.tty")

__all__ = ["Terminal"]

# =========================================================================
# Terminal class
# =========================================================================


class Terminal(object):

    """
    Terminal is used to bootstrap Junos New Out of the Box (NOOB) device
    over the CONSOLE port. The general use-case is to setup the minimal
    configuration so that the device is IP reachable using SSH
    and NETCONF for remote management.

    Serial is needed for Junos devices that do not support
    the DHCP 'auto-installation' or 'ZTP' feature; i.e. you *MUST*
    do the NOOB configuration via the CONSOLE.

    Serial is also useful for situations even when the Junos
    device supports auto-DHCP, but is not an option due to the
    specific situation
    """

    TIMEOUT = 0.2  # serial readline timeout, seconds
    EXPECT_TIMEOUT = 10  # total read timeout, seconds
    LOGIN_RETRY = 20  # total number of passes thru login state-machine

    _ST_INIT = 0
    _ST_LOADER = 1
    _ST_LOGIN = 2
    _ST_PASSWD = 3
    _ST_DONE = 4
    _ST_BAD_PASSWD = 5
    _ST_TTY_NOLOGIN = 6
    _ST_TTY_OPTION = 7
    _ST_TTY_HOTKEY = 8

    _re_pat_login = r"(?P<login>ogin:\s*$)"

    _RE_PAT = [
        r"(?P<loader>oader>\s*$)",
        _re_pat_login,
        r"(?P<passwd>assword:\s*$)",
        "(?P<badpasswd>ogin incorrect)",
        r"(?P<netconf_closed><!-- session end at .*-->\s*)",
        r"(?P<shell>%|#|(~\$)\s*$)",
        '(?P<cli>[^\\-"]>\s*$)',
        r"(?P<option>Enter your option:\s*$)",
        "(?P<hotkey>connection: <CTRL>Z)",
    ]

    # -----------------------------------------------------------------------
    # CONSTRUCTOR
    # -----------------------------------------------------------------------

    def __init__(self, **kvargs):
        """
        :kvargs['user']:
          defaults to 'root'

        :kvargs['passwd']:
          defaults to empty; NOOB Junos device there is
          no root password initially

        :kvargs['attempts']:
          the total number of login attempts thru the login
          state-machine
        """
        # logic args
        self.hostname = self.__dict__.get("host")
        self.user = kvargs.get("user", "root")
        self.passwd = kvargs.get("passwd", "")
        self.cs_user = kvargs.get("cs_user")
        self.cs_passwd = kvargs.get("cs_passwd")
        self.login_attempts = kvargs.get("attempts") or self.LOGIN_RETRY
        self.console_has_banner = kvargs.get("console_has_banner") or False
        self._huge_tree = kvargs.get("huge_tree", False)

        # misc setup
        self.nc = tty_netconf(self)
        self.state = self._ST_INIT
        self._badpasswd = 0
        self._loader = 0

    @property
    def tty_name(self):
        return self._tty_name

    # -----------------------------------------------------------------------
    # Login/logout
    # -----------------------------------------------------------------------

    def login(self):
        """
        open the TTY connection and login.  once the login is successful,
        start the NETCONF XML API process
        """
        logger.info("TTY: connecting to TTY:{} ...".format(self.tty_name))
        self._tty_open()

        logger.info("TTY: logging in......")

        self.state = self._ST_INIT
        self._login_state_machine()

        # now start NETCONF XML
        logger.info("TTY: OK.....starting NETCONF")
        self.nc.open(at_shell=self.at_shell)
        self.session_id = self.nc._session_id

        return True

    def logout(self):
        """
        cleanly logout of the TTY
        """
        logger.info("logout: logging out.....")
        self.nc.close()
        self._logout_state_machine()
        return True

        # ---------------------------------------------------------------------
        # TTY logout state-machine
        # ---------------------------------------------------------------------

    def _logout_state_machine(self, attempt=0):
        if 10 == attempt:
            raise RuntimeError("logout_sm_failure")

        prompt, found = self.read_prompt()

        def _ev_login():
            # back at login prompt, so we are cleanly done!
            self._tty_close()

        def _ev_shell():
            self.write("exit")

        def _ev_cli():
            self.write("exit")

        # Connection closed by foreign host
        def _ev_netconf_closed():
            return True

        _ev_tbl = {
            "login": _ev_login,
            "shell": _ev_shell,
            "cli": _ev_cli,
            "netconf_closed": _ev_netconf_closed,
        }

        # hack for now
        # in case of telnet to management port, after writing exit on console
        # it exits completely and returns None
        ###
        if found is not None:
            _ev_tbl[found]()
        else:
            return True

        if found == "login":
            return True

        else:
            sleep(1)
            return self._logout_state_machine(attempt=attempt + 1)

    # -----------------------------------------------------------------------
    # TTY login state-machine
    # -----------------------------------------------------------------------
    def _login_state_machine(self, attempt=0):
        if self.login_attempts == attempt:
            raise RuntimeError("login_sm_failure")

        prompt, found = self.read_prompt()

        def _ev_loader():
            self.state = self._ST_LOADER
            self.write("boot")
            self.write("\n")
            sleep(300)
            self._login_state_machine(attempt=0)
            self._loader += 1
            if self._loader == 2:
                raise RuntimeError("probably corrupted image, stuck in loader")

        def _ev_login():
            self.state = self._ST_LOGIN
            self.write(self.user)

        def _ev_passwd():
            self.state = self._ST_PASSWD
            self.write(self.passwd)

        def _ev_bad_passwd():
            self.state = self._ST_BAD_PASSWD
            self.write("\n")
            self._badpasswd += 1
            if self._badpasswd == 2:
                # raise RuntimeError("Bad username/password")
                raise EzErrors.ConnectAuthError(self, "Bad username/password")
            # return through and try again ... could have been
            # prior failed attempt

        def _ev_tty_nologin():
            if self._ST_INIT == self.state:
                # assume we're in a hung state, i.e. we don't see
                # a login prompt for whatever reason
                self.state = self._ST_TTY_NOLOGIN
                # For console based telnet connection a new-line is required.
                # Code modified to check with a newline for telnet based connections.
                # Keeping below code as a comment for future enhancement.
                # if self.console_has_banner:
                # # if console connection has a banner or warning,
                # # use this hack
                #     sleep(5)
                #     self.write("\n")
                sleep(5)
                self.write("\n")
            else:
                # @@@ this is still a hack - used by default
                self.write("<close-session/>")

        def _ev_shell():
            if self.state == self._ST_INIT:
                # this means that the shell was left
                # open.  probably not a good thing,
                # so issue a logging message, but move on.
                logger.warning("login_warn: Shell login was open!!")

            self.at_shell = True
            self.state = self._ST_DONE
            # if we are here, then we are done

        def _ev_cli():
            if self.state == self._ST_INIT:
                # this means that the shell was left open.  probably not a
                # good thing, so issue a logging message, hit <ENTER> and try
                # again just to be sure...
                logger.warning("login_warn: waiting on TTY..... ")
                sleep(5)
                #  return

            self.at_shell = False
            self.state = self._ST_DONE

        def _ev_option():
            self.state = self._ST_TTY_OPTION
            self.write("1")

        def _ev_hot_key():
            self.state = self._ST_TTY_HOTKEY
            self.write("\n")

        _ev_tbl = {
            "loader": _ev_loader,
            "login": _ev_login,
            "passwd": _ev_passwd,
            "badpasswd": _ev_bad_passwd,
            "shell": _ev_shell,
            "cli": _ev_cli,
            "option": _ev_option,
            "hotkey": _ev_hot_key,
        }

        _ev_tbl.get(found, _ev_tty_nologin)()

        if self.state == self._ST_DONE:
            return True
        else:
            # if we are here, then loop the event again
            self._login_state_machine(attempt + 1)



"""
FTP utility
"""

import re
import ftplib
import os
import logging

logger = logging.getLogger("jnpr.junos.utils.ftp")


class FTP(ftplib.FTP):
    """
    FTP utility can be used to transfer files to and from device.
    """

    def __init__(self, junos, **ftpargs):
        """

        :param Device junos: Device object
        :param kvargs ftpargs: any additional args to be passed to ftplib FTP

        Supports python *context-manager* pattern.  For example::

            from jnpr.junos.utils.ftp import FTP
            with FTP(dev) as ftp:
                ftp.put(package, remote_path)
        """

        self._junos = junos
        self._ftpargs = ftpargs
        ftplib.FTP.__init__(
            self,
            self._junos._hostname,
            self._junos._auth_user,
            self._junos._auth_password,
        )

    # dummy function, as connection is created by ftb lib in __init__ only
    def open(self):
        return self

    def put(self, local_file, remote_path=None):
        """
        This function is used to upload file to the router from local
        execution server/shell.

        :param local_file: Full path along with filename which has to be
            copied to router
        :param remote_path: path in which to receive the files on the remote
            host. If ignored FILE will be copied to "tmp"
        :returns: True if the transfer succeeds, else False

        """

        try:
            mat = re.search("^.*/(.*)$", local_file)
            if mat:
                if not remote_path:
                    remote_file = "/tmp/" + mat.group(1)
                else:
                    if re.search("^.*/(.*)$", remote_path) and re.search(
                        r"\.\w+$", remote_path
                    ):
                        remote_file = remote_path
                        # Looks like remote path is given as file location
                    else:
                        remote_file = os.path.join(remote_path, mat.group(1))
            else:
                if not remote_path:
                    remote_file = os.path.join("/tmp/", local_file)
                else:
                    remote_file = os.path.join(remote_path, local_file)

            with open(local_file, "rb") as open_local_file:
                self.storbinary(
                    cmd="STOR " + remote_file,
                    fp=open_local_file,
                    callback=self._ftpargs.get("callback"),
                )
        except Exception as ex:
            logger.error(ex)
            return False
        return True

    def get(self, remote_file, local_path=os.getcwd()):
        """
        This function is used to download file from router to local execution
        server/shell.

        :param local_path: path in which to receive files locally

        :param remote_file: Full path along with filename on the router. If
            ignored FILE will be copied to "tmp"

        :returns: True if the transfer succeeds, else False
        """
        if os.path.isdir(local_path):
            mat = re.search("^.*/(.*)$", remote_file)
            if mat:
                local_file = os.path.join(local_path, mat.group(1))
            else:
                local_file = local_path
        else:
            local_file = local_path

        with open(local_file, "wb") as local_fh:
            args_callback = self._ftpargs.get("callback")

            def callback(data):
                local_fh.write(data)
                if args_callback:
                    args_callback(data)

            try:
                self.retrbinary("RETR " + remote_file, callback)
            except Exception as ex:
                logger.error(ex)
                return False
            return True

    # -------------------------------------------------------------------------
    # CONTEXT MANAGER
    # -------------------------------------------------------------------------

    def __enter__(self):
        # return self.open(**self._ftpargs)
        return self

    def __exit__(self, exc_ty, exc_val, exc_tb):
        return self.close()



# stdlib
from copy import deepcopy

# local
from jnpr.junos.factory.cfgtable import CfgTable
from jnpr.junos.factory.optable import OpTable
from jnpr.junos.factory.cmdtable import CMDTable
from jnpr.junos.factory.table import Table

from jnpr.junos.factory.view import View
from jnpr.junos.factory.cmdview import CMDView

from jnpr.junos.utils.config import Config


def FactoryCfgTable(table_name=None, data_dict={}):
    if table_name is None:
        table_name = "CfgTable"
    if "set" in data_dict.keys():
        new_cls = type(table_name, (CfgTable, Config), {})
    else:
        new_cls = type(table_name, (CfgTable,), {})
    new_cls.DEFINE = deepcopy(data_dict)
    new_cls.__module__ = __name__.replace("factory_cls", "CfgTable")
    return new_cls


def FactoryOpTable(
    cmd,
    args=None,
    args_key=None,
    item=None,
    key=OpTable.ITEM_NAME_XPATH,
    view=None,
    table_name=None,
    use_filter=True,
):
    if table_name is None:
        table_name = "OpTable." + cmd
    new_cls = type(table_name, (OpTable,), {})
    new_cls.GET_RPC = cmd
    new_cls.GET_ARGS = args or {}
    if args_key is not None:
        new_cls.GET_KEY = args_key
    new_cls.ITEM_XPATH = item
    new_cls.ITEM_NAME_XPATH = key
    new_cls.VIEW = view
    new_cls.USE_FILTER = use_filter
    new_cls.__module__ = __name__.replace("factory_cls", "OpTable")
    return new_cls


def FactoryCMDTable(
    cmd,
    args=None,
    item=None,
    key_items=None,
    key="name",
    view=None,
    table_name=None,
    title=None,
    delimiter=None,
    eval=None,
    platform="juniper_junos",
    use_textfsm=False,
    **kwargs
):
    new_cls = type(table_name, (CMDTable,), {})
    new_cls.GET_CMD = cmd
    if "target" in kwargs:
        new_cls.TARGET = kwargs["target"]
    new_cls.KEY_ITEMS = key_items
    new_cls.CMD_ARGS = args or {}
    new_cls.ITEM = item
    new_cls.KEY = key
    new_cls.VIEW = view
    new_cls.TITLE = title
    new_cls.DELIMITER = delimiter
    new_cls.EVAL = eval
    new_cls.PLATFORM = platform
    new_cls.USE_TEXTFSM = use_textfsm
    new_cls.__module__ = __name__.replace("factory_cls", "CMDTable")
    return new_cls


def FactoryCMDChildTable(
    title=None,
    regex=None,
    key="name",
    delimiter=None,
    table_name=None,
    view=None,
    key_items=None,
    item=None,
    eval=None,
):
    new_cls = type(table_name, (CMDTable,), {})
    new_cls.DELIMITER = delimiter
    new_cls.KEY = key
    new_cls.REGEX = regex
    new_cls.TITLE = title
    new_cls.VIEW = view
    new_cls.KEY_ITEMS = key_items
    new_cls.ITEM = item
    new_cls.EVAL = eval
    new_cls.__module__ = __name__.replace("factory_cls", "CMDTable")
    return new_cls


def FactoryTable(
    item, key=Table.ITEM_NAME_XPATH, view=None, table_name=None, use_filter=True
):
    if table_name is None:
        table_name = "Table." + item
    new_cls = type(table_name, (Table,), {})
    new_cls.ITEM_XPATH = item
    new_cls.ITEM_NAME_XPATH = key
    new_cls.VIEW = view
    new_cls.USE_FILTER = use_filter
    new_cls.__module__ = __name__.replace("factory_cls", "Table")
    return new_cls


def FactoryView(fields, **kvargs):
    """
    :fields:
      dictionary of fields, structure of which is ~internal~ and should
      not be defined explicitly. use the RunstatMaker.Fields() mechanism to
      create theserather than hardcoding the dictionary structures;
      since they might change over time.

    :kvargs:
      'view_name' to name the class.  this could be useful for debug
      or eventual callback mechanisms.

      'groups' is a dict of name/xpath assocaited to fields
      this technique would be used to extract fields from
      node-set elements like port <if-device-flags>.

      'extends' names the base View class to extend.  using this
      technique you can add to existing defined Views.
    """

    view_name = kvargs.get("view_name", "RunstatView")
    new_cls = type(view_name, (View,), {})

    if "extends" in kvargs:
        base_cls = kvargs["extends"]
        new_cls.FIELDS = deepcopy(base_cls.FIELDS)
        new_cls.FIELDS.update(fields)
        if "groups" in kvargs:
            new_cls.GROUPS = deepcopy(base_cls.GROUPS)
            new_cls.GROUPS.update(kvargs["groups"])
    else:
        new_cls.FIELDS = fields
        new_cls.GROUPS = kvargs["groups"] if "groups" in kvargs else None

    if "eval" in kvargs:
        new_cls.EVAL = kvargs["eval"]
        new_cls.FIELDS.update(kvargs["eval"])

    new_cls.__module__ = __name__.replace("factory_cls", "View")
    return new_cls


def FactoryCMDView(fields, **kvargs):
    """
    :fields:
      dictionary of fields, structure of which is ~internal~ and should
      not be defined explicitly. use the RunstatMaker.Fields() mechanism to
      create theserather than hardcoding the dictionary structures;
      since they might change over time.

    :kvargs:
      'view_name' to name the class.  this could be useful for debug
      or eventual callback mechanisms.

      'groups' is a dict of name/xpath assocaited to fields
      this technique would be used to extract fields from
      node-set elements like port <if-device-flags>.

      'extends' names the base View class to extend.  using this
      technique you can add to existing defined Views.
    """

    view_name = kvargs.get("view_name", "RunstatView")
    new_cls = type(view_name, (CMDView,), {})

    if "columns" in kvargs:
        new_cls.COLUMNS = deepcopy(kvargs["columns"])
    elif "title" in kvargs:
        new_cls.TITLE = deepcopy(kvargs["title"])
    if "regex" in kvargs:
        new_cls.REGEX = deepcopy(kvargs["regex"])
    if "exists" in kvargs:
        new_cls.EXISTS = deepcopy(kvargs["exists"])
    if "filters" in kvargs:
        new_cls.FILTERS = deepcopy(kvargs["filters"])
    if fields is not None:
        new_cls.FIELDS = fields
    if "eval" in kvargs:
        new_cls.EVAL = kvargs["eval"]
        # new_cls.FIELDS.update(kvargs['eval'])

    new_cls.__module__ = __name__.replace("factory_cls", "CMDView")
    return new_cls



# 3rd-party
from lxml.builder import E

# local module
from jnpr.junos.cfg.resource import Resource


class PhyPortBase(Resource):

    """
    [edit interfaces <name>]

    Resource name: str
      <name> is the interface-name (IFD), e.g. 'ge-0/0/0'
    """

    PROPERTIES = [
        "admin",  # True
        "description",  # str
        "speed",  # ['10m','100m','1g','10g']
        "duplex",  # ['full','half']
        "mtu",  # int
        "loopback",  # True
        "$unit_count",  # number of units defined
    ]

    PORT_DUPLEX = {"full": "full-duplex", "half": "half-duplex"}

    @classmethod
    def _set_invert(cls, in_this, item, from_this):
        from_item = in_this[item]
        in_this[item] = [_k for _k, _v in from_this.items() if _v == from_item][0]

    # -----------------------------------------------------------------------
    # XML readers
    # -----------------------------------------------------------------------

    def _xml_at_top(self):
        return E.interfaces(E.interface(E.name(self._name)))

    def _xml_at_res(self, xml):
        return xml.find(".//interface")

    def _xml_to_py(self, has_xml, has_py):
        # common to all subclasses
        Resource._r_has_xml_status(has_xml, has_py)
        has_py["admin"] = bool(has_xml.find("disable") is None)
        Resource.copyifexists(has_xml, "description", has_py)
        Resource.copyifexists(has_xml, "mtu", has_py)
        has_py["$unit_count"] = len(has_xml.findall("unit"))

    # -----------------------------------------------------------------------
    # XML writers
    # -----------------------------------------------------------------------

    # description handed by Resource

    def _xml_change_admin(self, xml):
        xml.append(Resource.xmltag_set_or_del("disable", (self.admin is False)))
        return True

    def _xml_change_mtu(self, xml):
        Resource.xml_set_or_delete(xml, "mtu", self.mtu)
        return True

    # -----------------------------------------------------------------------
    # Manager List, Catalog
    # -----------------------------------------------------------------------

    def _r_list(self):
        got = self.R.get_interface_information(
            media=True, interface_name="[efgx][et]-*"
        )
        self._rlist = [
            name.text.strip() for name in got.xpath("physical-interface/name")
        ]



def facts_srx_cluster(junos, facts):
    # we should check the 'cluster status' on redundancy group 0 to see who is
    # master.  we use a try/except block for cases when SRX is not clustered

    try:
        cluster_st = junos.rpc.get_chassis_cluster_status(redundancy_group="0")
        if "error" == cluster_st.tag:
            facts["srx_cluster"] = False
            return

        primary = cluster_st.xpath('.//redundancy-group-status[.="primary"]')[0]

        node = primary.xpath("preceding-sibling::device-name[1]")[0].text.replace(
            "node", "RE"
        )

        if not facts.get("master"):
            facts["master"] = node
        elif node not in facts["master"]:
            facts["master"].append(node)

        facts["srx_cluster"] = True

    except:
        # this device doesn't support SRX chassis clustering; i.e.
        # since we arbitrarily execute the RPC on all devices, if we
        # hit this exception we just ignore, A-OK, yo!
        pass



import re


def provides_facts():
    """
    Returns a dictionary keyed on the facts provided by this module. The value
    of each key is the doc string describing the fact.
    """
    return {
        "personality": "A string which is generally based on the "
        "platform and indicates the behavior of the "
        "device.",
        "virtual": "A boolean indicating if the device is virtual.",
    }


def get_facts(device):
    """
    Determines personality fact based on the model.
    """
    personality = None
    virtual = None

    model = device.facts["model"]

    if model == "Virtual Chassis":
        # Set model to the model of the first RE in the multi-chassis system.
        model = device.facts["re_info"]["default"]["default"]["model"]

    if re.match("^(EX)|(QFX)", model):
        personality = "SWITCH"
        virtual = False
    elif model.startswith("MX"):
        re_type = device.facts["re_info"]["default"]["default"]["model"]
        # The VMX has an RE type of 'RE-VMX'
        if re_type == "RE-VMX":
            personality = "MX"
            virtual = True
        # An MX GNF has an RE type that includes the letters 'GNF'
        elif re_type is not None and "GNF" in re_type:
            personality = "MX-GNF"
            virtual = True
        else:
            personality = "MX"
            virtual = False
    elif model.startswith("VMX"):
        personality = "MX"
        virtual = True
    elif model.startswith("VJX"):
        personality = "SRX_BRANCH"
        virtual = True
    elif "VRR" == model:
        personality = "MX"
        virtual = True
    elif model.startswith("M"):
        personality = "M"
        virtual = False
    elif model.startswith("T"):
        personality = "T"
        virtual = False
    elif model.startswith("PTX"):
        personality = "PTX"
        # The vPTX has an RE type of 'RE-VIRTUAL'
        if device.facts["re_info"]["default"]["default"]["model"] == "RE-VIRTUAL":
            virtual = True
        else:
            virtual = False
    elif re.match(r"SRX\s?(\d){4}", model):
        srx_model = int(model[-4:])
        if srx_model > 5000:
            personality = "SRX_HIGHEND"
        else:
            personality = "SRX_MIDRANGE"
        virtual = False
    elif re.match(r"SRX\s?(\d){3}", model):
        personality = "SRX_BRANCH"
        virtual = False
    elif re.search("firefly", model, re.IGNORECASE):
        personality = "SRX_BRANCH"
        virtual = True
    elif "OLIVE" == model:
        personality = "OLIVE"
        virtual = True
    elif model.startswith("NFX"):
        personality = "NFX"
        virtual = False
    elif "JUNOS_NODE_SLICING" == model:
        personality = "JDM"
        virtual = True
    elif model.startswith("ACX"):
        personality = "ACX"
        virtual = False

    return {
        "personality": personality,
        "virtual": virtual,
    }
